# coding: utf-8
"""
Add NOT NULL constraint for data unless fed_id and component_id are not null.
"""

import flask_sqlalchemy


def run(db: flask_sqlalchemy.SQLAlchemy) -> bool:
    constraints = db.session.execute(db.text("""
         SELECT conname
         FROM pg_catalog.pg_constraint
         WHERE conname = 'files_not_null_check_data'
    """)).fetchall()

    if len(constraints) > 0:
        return False

    db.session.execute(db.text("""
            UPDATE files SET data = '{"storage": "local", "original_file_name": ""}'::json WHERE data IS NULL AND (fed_id IS NULL OR component_id IS NULL)
        """))

    # Perform migration
    db.session.execute(db.text("""
        ALTER TABLE files
        ADD CONSTRAINT files_not_null_check_data
            CHECK (
                (
                    fed_id IS NOT NULL AND
                    component_id IS NOT NULL
                ) OR data IS NOT NULL
            )
        """))
    return True



# coding: utf-8
"""

"""

import enum
import datetime
import typing

from sqlalchemy.orm import Mapped, Query

from .. import db
from .objects import Objects
from .users import User
from .utils import Model


@enum.unique
class DataverseExportStatus(enum.Enum):
    TASK_CREATED = 0
    EXPORT_FINISHED = 1


class DataverseExport(Model):
    __tablename__ = 'dataverse_exports'

    object_id: Mapped[int] = db.Column(db.Integer, db.ForeignKey(Objects.object_id_column), primary_key=True)
    dataverse_url: Mapped[typing.Optional[str]] = db.Column(db.String, nullable=True)
    user_id: Mapped[int] = db.Column(db.Integer, db.ForeignKey(User.id), nullable=False)
    utc_datetime: Mapped[datetime.datetime] = db.Column(db.TIMESTAMP(timezone=True), nullable=False)
    status: Mapped[DataverseExportStatus] = db.Column(db.Enum(DataverseExportStatus), nullable=False)

    if typing.TYPE_CHECKING:
        query: typing.ClassVar[Query["DataverseExport"]]

    def __init__(
            self,
            object_id: int,
            dataverse_url: typing.Optional[str],
            user_id: int,
            status: DataverseExportStatus,
            utc_datetime: typing.Optional[datetime.datetime] = None
    ) -> None:
        super().__init__(
            object_id=object_id,
            dataverse_url=dataverse_url,
            user_id=user_id,
            utc_datetime=utc_datetime if utc_datetime is not None else datetime.datetime.now(datetime.timezone.utc),
            status=status
        )

    def __repr__(self) -> str:
        return f'<{type(self).__name__}(object_id={self.object_id}, dataverse_url={self.dataverse_url})>'



# coding: utf-8
"""

"""
import typing

from sqlalchemy.dialects import postgresql
from sqlalchemy.orm import Mapped, Query

from .. import db
from .utils import Model


class MarkdownToHTMLCacheEntry(Model):
    __tablename__ = 'markdown_to_html_cache_entries'

    id: Mapped[int] = db.Column(db.Integer, primary_key=True)
    markdown: Mapped[str] = db.Column(db.Text, nullable=False)
    parameters: Mapped[typing.Dict[str, typing.Any]] = db.Column(postgresql.JSONB, nullable=False)
    html: Mapped[str] = db.Column(db.Text, nullable=False)

    if typing.TYPE_CHECKING:
        query: typing.ClassVar[Query["MarkdownToHTMLCacheEntry"]]



# coding: utf-8
"""

"""
from __future__ import annotations

import dataclasses
import datetime
import itertools
import typing
import sys
import operator

import flask

from .actions import get_action, Action
from .files import get_files_for_object, File
from .objects import get_object, find_object_references
from .object_permissions import get_user_permissions_for_multiple_objects, get_objects_with_permissions
from ..models import ObjectLogEntryType, ObjectLogEntry, Permissions, Object
from .. import db

# limit tree depth to avoid hitting the recursion limit in the frontend
_MAXIMUM_TREE_DEPTH = sys.getrecursionlimit() - 100


def get_related_object_ids(
        object_id: int,
        include_referenced_objects: bool,
        include_referencing_objects: bool,
        user_id: typing.Optional[int] = None,
        filter_referencing_objects_by_permissions: bool = True
) -> typing.Tuple[
    typing.List[typing.Tuple[int, typing.Optional[str]]], typing.List[typing.Tuple[int, typing.Optional[str]]]
]:
    """
    Get IDs of objects related to a given object.

    An object is considered to be related, if it is:
    - a previous object of the given object (include_referenced_objects)
    - a measurement created using the given object (include_referencing_objects)
    - a sample created using the given object (include_referencing_objects)
    - any object referencing the given object in its metadata (include_referencing_objects)

    :param object_id: the ID of an existing object
    :param include_referenced_objects: whether to include referenced objects
    :param include_referencing_objects: whether to include referencing objects
    :param user_id: the ID of an existing user (optional)
    :param filter_referencing_objects_by_permissions: whether the referencing
        object ID list should only contain objects for which the given user has
        READ permissions
    :return: lists of previous object IDs, measurement IDs and sample IDs
    """
    referenced_object_ids: typing.Set[typing.Tuple[int, typing.Optional[str]]] = set()
    referencing_object_ids: typing.Set[typing.Tuple[int, typing.Optional[str]]] = set()
    if include_referenced_objects:
        object = get_object(object_id)
        for referenced_object_id, _previously_referenced_object_id, _schema_type in find_object_references(object_id=object.object_id, version_id=object.version_id, object_data=object.data, include_fed_references=True):
            if referenced_object_id is not None:
                referenced_object_ids.add(referenced_object_id)
    if include_referencing_objects:
        object_log_entries = ObjectLogEntry.query.filter(
            db.and_(
                ObjectLogEntry.object_id == object_id,
                ObjectLogEntry.type.in_((
                    ObjectLogEntryType.USE_OBJECT_IN_MEASUREMENT,
                    ObjectLogEntryType.USE_OBJECT_IN_SAMPLE_CREATION,
                    ObjectLogEntryType.REFERENCE_OBJECT_IN_METADATA
                ))
            )
        ).all()
        object_id_key_map = {
            ObjectLogEntryType.USE_OBJECT_IN_MEASUREMENT: 'measurement_id',
            ObjectLogEntryType.USE_OBJECT_IN_SAMPLE_CREATION: 'sample_id',
            ObjectLogEntryType.REFERENCE_OBJECT_IN_METADATA: 'object_id',
        }
        for object_log_entry in object_log_entries:
            object_id_key = object_id_key_map.get(object_log_entry.type, 'object_id')
            referencing_object_id = object_log_entry.data.get(object_id_key)
            if referencing_object_id is not None:
                referencing_object_ids.add((referencing_object_id, None))
        if filter_referencing_objects_by_permissions:
            object_permissions = get_user_permissions_for_multiple_objects(user_id, [referencing_object_id[0] for referencing_object_id in referencing_object_ids])
            referencing_object_ids = {
                referencing_object_id
                for referencing_object_id in referencing_object_ids
                if Permissions.READ in object_permissions[referencing_object_id[0]]
            }
    return list(referencing_object_ids), list(referenced_object_ids)


@dataclasses.dataclass(kw_only=True, frozen=True)
class ObjectRef:
    object_id: int
    component_uuid: typing.Optional[str]
    eln_source_url: typing.Optional[str]
    eln_object_url: typing.Optional[str]

    @property
    def is_local(self) -> bool:
        return (self.component_uuid is None or self.component_uuid == flask.current_app.config['FEDERATION_UUID']) and self.eln_source_url is None and self.eln_object_url is None


@dataclasses.dataclass(kw_only=True)
class RelatedObjectsSubTree:
    referenced_objects: typing.List[ObjectRef]
    referencing_objects: typing.List[ObjectRef]


@dataclasses.dataclass(kw_only=True)
class RelatedObjectsTree:
    object_ref: ObjectRef
    path: typing.List[typing.Union[int, ObjectRef]]
    object: typing.Optional[Object]
    object_name: typing.Optional[str]
    referenced_objects: typing.Optional[typing.List[RelatedObjectsTree]]
    referencing_objects: typing.Optional[typing.List[RelatedObjectsTree]]

    @property
    def object_id(self) -> int:
        return self.object_ref.object_id

    @property
    def component_uuid(self) -> typing.Optional[str]:
        return self.object_ref.component_uuid

    @property
    def eln_source_url(self) -> typing.Optional[str]:
        return self.object_ref.eln_source_url

    @property
    def eln_object_url(self) -> typing.Optional[str]:
        return self.object_ref.eln_object_url


def get_referencing_object_ids(
        object_ids: typing.Set[int]
) -> typing.Dict[int, typing.Set[ObjectRef]]:
    if not object_ids:
        return {}
    referencing_object_ids_by_id: typing.Dict[int, typing.Set[ObjectRef]] = {
        object_id: set()
        for object_id in object_ids
    }
    for object_id, referencing_object_id, referencing_component_uuid, referencing_eln_source_url, referencing_eln_object_url in db.session.execute(db.text("""
        SELECT DISTINCT object_id, coalesce(data -> 'object_id', data -> 'sample_id', data -> 'measurement_id', NULL)::text::integer AS referencing_object_id, NULL AS referencing_component_uuid, NULL AS eln_source_url, NULL AS eln_object_url
        FROM object_log_entries
        WHERE object_id IN :object_ids AND type IN ('USE_OBJECT_IN_MEASUREMENT', 'USE_OBJECT_IN_SAMPLE_CREATION', 'REFERENCE_OBJECT_IN_METADATA')
    """), {'object_ids': tuple(object_ids)}).fetchall():
        referencing_object_ids_by_id[object_id].add(ObjectRef(object_id=referencing_object_id, component_uuid=referencing_component_uuid, eln_source_url=referencing_eln_source_url, eln_object_url=referencing_eln_object_url))
    return referencing_object_ids_by_id


def _get_referenced_object_ids(
        object_ids: typing.Set[int]
) -> typing.Dict[int, typing.Set[ObjectRef]]:
    if not object_ids:
        return {}
    referenced_object_ids_by_id: typing.Dict[int, typing.Set[ObjectRef]] = {
        object_id: set()
        for object_id in object_ids
    }
    for object_id, referenced_object_id, referenced_component_uuid, referenced_eln_source_url, referenced_eln_object_url in db.session.execute(db.text("""
        WITH RECURSIVE properties(object_id, value) AS (
            SELECT objects_current.object_id, data.value
            FROM objects_current, jsonb_each(objects_current.data) AS data
            WHERE objects_current.object_id IN :object_ids AND jsonb_typeof(objects_current.data) = 'object'
        UNION ALL
            SELECT properties.object_id, coalesce(object_data.value, array_data.value)
            FROM properties
            LEFT OUTER JOIN jsonb_each(CASE
                WHEN jsonb_typeof(properties.value) = 'object' THEN properties.value
                ELSE '{}'::jsonb
            END) AS object_data ON jsonb_typeof(properties.value) = 'object'
            LEFT OUTER JOIN jsonb_array_elements(CASE
                WHEN jsonb_typeof(properties.value) = 'array' THEN properties.value
                ELSE '[]'::jsonb
            END) AS array_data ON jsonb_typeof(properties.value) = 'array'
            WHERE jsonb_typeof(properties.value) = 'object' OR jsonb_typeof(properties.value) = 'array'
        )
        SELECT object_id, (properties.value -> 'object_id')::integer, properties.value ->> 'component_uuid', properties.value ->> 'eln_source_url', properties.value ->> 'eln_object_url'
        FROM properties
        WHERE jsonb_typeof(properties.value) = 'object' AND properties.value ? '_type' AND properties.value ? 'object_id' AND properties.value ->> '_type' IN ('object_reference', 'sample', 'measurement')
    """), {'object_ids': tuple(object_ids)}).fetchall():
        referenced_object_ids_by_id[object_id].add(ObjectRef(object_id=referenced_object_id, component_uuid=referenced_component_uuid, eln_source_url=referenced_eln_source_url, eln_object_url=referenced_eln_object_url))
    return referenced_object_ids_by_id


def gather_related_object_subtrees(
        object_id: int
) -> typing.Dict[ObjectRef, RelatedObjectsSubTree]:
    object_ref = ObjectRef(
        object_id=object_id,
        component_uuid=None,
        eln_source_url=None,
        eln_object_url=None,
    )
    subtrees: typing.Dict[ObjectRef, RelatedObjectsSubTree] = {}
    object_ref_stack: typing.List[typing.Tuple[ObjectRef, typing.Optional[ObjectRef]]] = [
        (object_ref, None)
    ]
    referencing_object_ids_by_id: typing.Dict[int, typing.Set[ObjectRef]] = {}
    referenced_object_ids_by_id: typing.Dict[int, typing.Set[ObjectRef]] = {}
    while object_ref_stack:
        object_ref, parent_object_ref = object_ref_stack.pop()
        if object_ref not in subtrees:
            subtrees[object_ref] = RelatedObjectsSubTree(referenced_objects=[], referencing_objects=[])

            if object_ref.is_local:
                referencing_object_ids = referencing_object_ids_by_id.get(object_ref.object_id)
                referenced_object_ids = referenced_object_ids_by_id.get(object_ref.object_id)
                if referencing_object_ids is None:
                    referencing_object_ids = get_referencing_object_ids({object_ref.object_id})[object_ref.object_id]
                if referenced_object_ids is None:
                    referenced_object_ids = _get_referenced_object_ids({object_ref.object_id})[object_ref.object_id]

                for child_object_list, filtered_child_object_list in [
                    (referenced_object_ids, subtrees[object_ref].referenced_objects),
                    (referencing_object_ids, subtrees[object_ref].referencing_objects)
                ]:
                    for child_object_ref in child_object_list:
                        if child_object_ref != parent_object_ref:
                            filtered_child_object_list.append(child_object_ref)
                            object_ref_stack.append((child_object_ref, object_ref))
                local_child_object_ids = {
                    child_object_ref.object_id
                    for child_object_ref in itertools.chain(subtrees[object_ref].referenced_objects, subtrees[object_ref].referencing_objects)
                    if child_object_ref.is_local and child_object_ref not in subtrees
                }
                referencing_object_ids_by_id.update(get_referencing_object_ids(local_child_object_ids))
                referenced_object_ids_by_id.update(_get_referenced_object_ids(local_child_object_ids))
    return subtrees


@dataclasses.dataclass(frozen=True)
class WorkflowElement:
    object_id: int
    object: typing.Optional[Object]
    action: typing.Optional[Action]
    is_referencing: bool
    is_referenced: bool
    files: typing.List[File]
    is_current_referenced: bool = True
    is_current_referencing: bool = True
    path: typing.Optional[typing.List[WorkflowElement]] = None
    duplicate: bool = False


@dataclasses.dataclass(frozen=True)
class ActionFilter:
    action_ids: typing.Optional[typing.List[int]] = None
    action_type_ids: typing.Optional[typing.List[int]] = None
    filter_operator: typing.Callable[[bool, bool], bool] = operator.and_

    @property
    def is_none(self) -> bool:
        return self.action_ids is None and self.action_type_ids is None

    def __call__(self, action: typing.Optional[Action]) -> bool:
        return self.filter_operator(
            self.action_ids is None or
            (
                action is not None and
                action.id in self.action_ids
            ),
            self.action_type_ids is None or
            (
                action is not None and
                action.type is not None and
                (
                    action.type.id in self.action_type_ids or
                    (
                        action.type.fed_id is not None and
                        action.type.fed_id < 0 and
                        action.type.fed_id in self.action_type_ids
                    )
                )
            )
        )


@dataclasses.dataclass(frozen=True)
class ActionFilters:
    referencing_objects_filter: ActionFilter = ActionFilter()
    referenced_objects_filter: ActionFilter = ActionFilter()

    @property
    def is_none(self) -> bool:
        return self.referencing_objects_filter.is_none and self.referenced_objects_filter.is_none


def get_workflow_references(object: Object, user_id: int, actions_by_id: typing.Optional[typing.Dict[int, Action]] = None) -> typing.List[typing.List[WorkflowElement]]:
    """
    Creates a list describing all direct object relations for workflow views as configured in an object's schema.

    :param object: The object for which workflow views are to be created
    :param user_id: User ID of the user viewing the workflow views taking into account the access permissions
    :param actions_by_id: A dict containing cached actions by ID
    :return: List of lists of WorkflowElements containing, object ID, object data, action data and if the object is referenced
            and/or referencing the object the view is created for
    """
    if object.schema is None or ('workflow_view' not in object.schema.keys() and 'workflow_views' not in object.schema.keys()):
        return []

    if 'workflow_view' in object.schema:
        workflow_views = [object.schema['workflow_view']]
    else:
        workflow_views = object.schema['workflow_views']

    if actions_by_id is None:
        actions_by_id = {}

    initial_object_version_by_id = {}
    objects_by_id: typing.Dict[int, Object] = {}
    files_by_object_id: typing.Dict[int, typing.List[File]] = {}

    workflows: typing.List[typing.List[WorkflowElement]] = [
        []
        for _ in workflow_views
    ]

    def _handle_object(
        object: Object,
        workflow_index: int,
        path: typing.List[WorkflowElement],
        actions_by_id: typing.Dict[int, Action],
        workflow_action_filters: ActionFilters,
        recursion_action_filters: ActionFilters,
        max_depth: typing.Optional[int]
    ) -> None:
        if path is None:
            path = []

        wf = workflow_action_filters
        rf = recursion_action_filters

        referencing_object_ids = {
            object_ref.object_id
            for object_ref in get_referencing_object_ids({object.object_id})[object.object_id]
            if object_ref.is_local
        }

        referencing_objects_current = {}
        for referencing_object_id in referencing_object_ids:
            referencing_objects_current[referencing_object_id] = False
            referencing_object = get_object(referencing_object_id)
            for referenced_object_id, _previously_referenced_object_id, _schema_type in find_object_references(
                    object_id=referencing_object.object_id,
                    version_id=referencing_object.version_id,
                    object_data=referencing_object.data
            ):
                if referenced_object_id == object.object_id:
                    referencing_objects_current[referencing_object_id] = True
                    break

        referenced_object_ids = {
            object_ref.object_id
            for object_ref in _get_referenced_object_ids({object.object_id})[object.object_id]
            if object_ref.is_local
        }

        object_ids = list(referencing_object_ids.union(referenced_object_ids))

        new_object_ids = list(set(object_ids).difference(set(objects_by_id.keys())))
        if new_object_ids:
            new_objects = get_objects_with_permissions(user_id, Permissions.READ, object_ids=new_object_ids)
        else:
            new_objects = []

        objects_by_id.update({
            object.object_id: object
            for object in new_objects
        })

        for object_id in new_object_ids:
            current_object_version = objects_by_id.get(object_id)
            if current_object_version is not None and current_object_version.version_id == 0:
                initial_object_version_by_id[object_id] = current_object_version
            else:
                initial_object_version_by_id[object_id] = get_object(object_id, version_id=0)

        for object_id in object_ids:
            action_id = initial_object_version_by_id[object_id].action_id
            action = None
            if action_id:
                if action_id in actions_by_id:
                    action = actions_by_id[action_id]
                else:
                    action = get_action(action_id)
                    actions_by_id[action_id] = action

            if (
                wf.is_none or
                (
                    object_id in referencing_object_ids and wf.referencing_objects_filter(action)
                ) or (
                    object_id in referenced_object_ids and wf.referenced_objects_filter(action)
                )
            ):
                if object_id in objects_by_id:
                    if object_id not in files_by_object_id:
                        files_by_object_id[object_id] = get_files_for_object(object_id)
                    files = files_by_object_id[object_id]
                else:
                    files = []
                elem_object = objects_by_id.get(object_id)
                welem = WorkflowElement(
                    object_id=object_id,
                    object=elem_object,
                    action=action,
                    is_referenced=object_id in referenced_object_ids,
                    is_referencing=object_id in referencing_object_ids,
                    files=files,
                    is_current_referenced=object_id in referenced_object_ids,
                    is_current_referencing=referencing_objects_current.get(object_id, False),
                    path=path,
                    duplicate=object_id in objects_in_workflow
                )
                objects_in_workflow.add(object_id)
                workflows[workflow_index].append(welem)

                if ((welem.is_referenced and welem.is_current_referenced) or (welem.is_referencing and welem.is_current_referencing)) and elem_object and (max_depth is None or len(path) < max_depth) and not welem.duplicate and (
                    rf.is_none or (
                        object_id in referenced_object_ids and elem_object.action_id and rf.referenced_objects_filter(get_action(elem_object.action_id))
                    ) or (
                        object_id in referencing_object_ids and elem_object.action_id and rf.referencing_objects_filter(get_action(elem_object.action_id))
                    )
                ):
                    _handle_object(elem_object, workflow_index, path + [welem], actions_by_id, workflow_action_filters, recursion_action_filters, max_depth)

    def _get_action_filter_lists(dictionary: typing.Dict[str, typing.Any], default_filter_operator: typing.Callable[[bool, bool], bool]) -> ActionFilters:
        return ActionFilters(
            referenced_objects_filter=ActionFilter(
                action_ids=[dictionary['referenced_action_id']] if isinstance(dictionary.get('referenced_action_id'), int) else dictionary.get('referenced_action_id'),
                action_type_ids=[dictionary['referenced_action_type_id']] if isinstance(dictionary.get('referenced_action_type_id'), int) else dictionary.get('referenced_action_type_id'),
                filter_operator={'or': operator.or_, 'and': operator.and_}.get(typing.cast(str, dictionary.get('referenced_filter_operator', '')), default_filter_operator)
            ),
            referencing_objects_filter=ActionFilter(
                action_ids=[dictionary['referencing_action_id']] if isinstance(dictionary.get('referencing_action_id'), int) else dictionary.get('referencing_action_id'),
                action_type_ids=[dictionary['referencing_action_type_id']] if isinstance(dictionary.get('referencing_action_type_id'), int) else dictionary.get('referencing_action_type_id'),
                filter_operator={'or': operator.or_, 'and': operator.and_}.get(typing.cast(str, dictionary.get('referencing_filter_operator', '')), default_filter_operator)
            )
        )

    for workflow_index, workflow_view in enumerate(workflow_views):
        objects_in_workflow = {object.object_id}

        workflow_action_filters = _get_action_filter_lists(workflow_view, operator.and_)

        recursion_filters = workflow_view.get('recursion_filters')
        if recursion_filters:
            max_depth = recursion_filters.get('max_depth', None)
            recursion_action_filters = _get_action_filter_lists(recursion_filters, operator.or_)
        else:
            max_depth = 0  # No filters defined -> no recursion
            recursion_action_filters = ActionFilters()

        _handle_object(object, workflow_index, [], actions_by_id, workflow_action_filters, recursion_action_filters, max_depth)

    # sorting key to sort by datetime
    def creation_time_key(workflow_element: WorkflowElement) -> datetime.datetime:
        object_id = workflow_element.object_id
        current_object_version = objects_by_id.get(object_id)
        if current_object_version is not None and current_object_version.data is not None:
            for datetime_attribute in workflow_views[0].get('sorting_properties', []):
                if isinstance(current_object_version.data.get(datetime_attribute), dict) and current_object_version.data[datetime_attribute].get('_type') == 'datetime':
                    return datetime.datetime.strptime(current_object_version.data[datetime_attribute]['utc_datetime'], '%Y-%m-%d %H:%M:%S').replace(tzinfo=datetime.timezone.utc)
        initital_object_version = initial_object_version_by_id.get(object_id)
        if initital_object_version is None or initital_object_version.utc_datetime is None:
            return datetime.datetime.max.replace(tzinfo=datetime.timezone.utc)
        return initital_object_version.utc_datetime

    # sort by the sorting key
    for workflow in workflows:
        workflow.sort(key=creation_time_key)

    return workflows



# coding: utf-8
"""
Replace columns of type TIMESTAMP WITHOUT TIME ZONE with columns of type
TIMESTAMP WITH TIME ZONE, using UTC for the conversion.
"""

import flask_sqlalchemy

from .utils import timestamp_add_timezone_utc_migration


def run(db: flask_sqlalchemy.SQLAlchemy) -> bool:
    migrations_ran = [
        timestamp_add_timezone_utc_migration(table_name, column_name)
        for table_name, column_name in [
            ('api_log_entries', 'utc_datetime'),
            ('background_tasks', 'expiration_date'),
            ('comments', 'utc_datetime'),
            ('components', 'last_sync_timestamp'),
            ('dataverse_exports', 'utc_datetime'),
            ('download_service_job_files', 'creation'),
            ('download_service_job_files', 'expiration'),
            ('eln_imports', 'import_utc_datetime'),
            ('eln_imports', 'upload_utc_datetime'),
            ('fed_action_log_entries', 'utc_datetime'),
            ('fed_action_type_log_entries', 'utc_datetime'),
            ('fed_comment_log_entries', 'utc_datetime'),
            ('fed_file_log_entries', 'utc_datetime'),
            ('fed_instrument_log_entries', 'utc_datetime'),
            ('fed_location_log_entries', 'utc_datetime'),
            ('fed_location_type_log_entries', 'utc_datetime'),
            ('fed_object_location_assignment_log_entries', 'utc_datetime'),
            ('fed_object_log_entries', 'utc_datetime'),
            ('fed_user_aliases', 'last_modified'),
            ('fed_user_log_entries', 'utc_datetime'),
            ('file_log_entries', 'utc_datetime'),
            ('files', 'utc_datetime'),
            ('group_invitations', 'utc_datetime'),
            ('instrument_log_entry_versions', 'event_utc_datetime'),
            ('instrument_log_entry_versions', 'utc_datetime'),
            ('location_log_entries', 'utc_datetime'),
            ('markdown_images', 'utc_datetime'),
            ('notifications', 'utc_datetime'),
            ('object_location_assignments', 'utc_datetime'),
            ('object_log_entries', 'utc_datetime'),
            ('object_shares', 'utc_datetime'),
            ('objects_current', 'utc_datetime'),
            ('objects_previous', 'utc_datetime'),
            ('objects_subversions', 'utc_datetime'),
            ('objects_subversions', 'utc_datetime_subversion'),
            ('project_invitations', 'utc_datetime'),
            ('scicat_exports', 'utc_datetime'),
            ('temporary_files', 'utc_datetime'),
            ('user_invitations', 'utc_datetime'),
            ('user_log_entries', 'utc_datetime'),
            ('users', 'last_modified')
        ]
    ]
    return any(migrations_ran)



# coding: utf-8
"""

"""

import typing
from .generate_placeholder import generate_placeholder


def copy_data(
        instance: typing.Union[typing.Dict[str, typing.Any], typing.List[typing.Any]],
        schema: typing.Dict[str, typing.Any]
) -> typing.Optional[typing.Union[typing.Dict[str, typing.Any], typing.List[typing.Any]]]:
    if not schema.get('may_copy', True):
        return generate_placeholder(schema)
    if schema['type'] == 'file':
        return generate_placeholder(schema)
    if schema['type'] == 'array' and isinstance(instance, list):
        array_copy: typing.List[typing.Any] = []
        for item in instance:
            array_copy.append(copy_data(item, schema['items']))
        return array_copy
    if schema['type'] == 'object' and isinstance(instance, dict):
        object_copy: typing.Dict[str, typing.Any] = {}
        for property_name, property_schema in schema['properties'].items():
            if property_name in instance:
                object_copy[property_name] = copy_data(instance[property_name], property_schema)
            elif not property_schema.get('may_copy', True):
                object_copy[property_name] = generate_placeholder(property_schema)
        return object_copy
    return instance



import typing
import datetime

import flask

from .components import _get_or_create_component_id
from .utils import _get_id, _get_uuid, _get_utc_datetime, _get_str, _get_dict
from .users import _parse_user_ref, _get_or_create_user_id, UserRef
from ..files import get_mutable_file, create_fed_file, hide_file, File
from ..components import Component
from .. import errors, fed_logs, object_log
from ..utils import parse_url
from ...models import Object
from ... import db


class FileHideData(typing.TypedDict):
    user: UserRef
    reason: str
    utc_datetime: datetime.datetime


class FileData(typing.TypedDict):
    fed_id: int
    component_uuid: str
    user: typing.Optional[UserRef]
    data: typing.Optional[typing.Dict[str, typing.Any]]
    utc_datetime: datetime.datetime
    hide: typing.Optional[FileHideData]


def parse_file(
        file_data: typing.Dict[str, typing.Any]
) -> FileData:
    uuid = _get_uuid(file_data.get('component_uuid'))
    if uuid == flask.current_app.config['FEDERATION_UUID']:
        # do not accept updates for own data
        raise errors.InvalidDataExportError('Invalid update for local data')
    fed_id = _get_id(file_data.get('file_id'), min=0)
    data = _get_dict(file_data.get('data'))

    if data:
        storage = _get_str(data.get('storage'), mandatory=True)
        _get_str(data.get('original_file_name'))
        if storage == 'url':
            for key in data.keys():
                if key not in {'url', 'storage'}:
                    raise errors.InvalidDataExportError(f'Invalid data entry key \'{key}\' for file #{fed_id} @ {uuid}')
            url = _get_str(data.get('url'))
            if url:
                try:
                    parse_url(url)
                except errors.InvalidURLError:
                    raise errors.InvalidDataExportError(f'Invalid URL \'{url}\' for file #{fed_id} @ {uuid}')
                except errors.URLTooLongError:
                    raise errors.InvalidDataExportError(f'URL \'{url}\' is too long for file #{fed_id} @ {uuid}')
                except errors.InvalidIPAddressError:
                    raise errors.InvalidDataExportError(f'Invalid IP-address in URL \'{url}\' for file #{fed_id} @ {uuid}')
                except errors.InvalidPortNumberError:
                    raise errors.InvalidDataExportError(f'Invalid port number in URL \'{url}\' for file #{fed_id} @ {uuid}')
        elif storage == 'federation':
            for key in data.keys():
                if key not in {'original_file_name', 'storage', 'hash'}:
                    raise errors.InvalidDataExportError(f'Invalid data entry key \'{key}\' for file #{fed_id} @ {uuid}')
        else:
            raise errors.InvalidDataExportError(f'Invalid storage type \'{storage}\' for file #{fed_id} @ {uuid}')

    hidden_data = _get_dict(file_data.get('hidden'), default=None)
    if data is None and hidden_data is None:
        raise errors.InvalidDataExportError(f'Missing data for file #{fed_id} @ {uuid}')

    if hidden_data is not None:
        hide = FileHideData(
            user=_parse_user_ref(_get_dict(hidden_data.get('user'), mandatory=True)),
            reason=_get_str(hidden_data.get('reason'), default=''),
            utc_datetime=_get_utc_datetime(hidden_data.get('utc_datetime'), default=datetime.datetime.now(datetime.timezone.utc))
        )
    else:
        hide = None

    return FileData(
        fed_id=fed_id,
        component_uuid=_get_uuid(file_data.get('component_uuid')),
        user=_parse_user_ref(_get_dict(file_data.get('user'))),
        data=data,
        utc_datetime=_get_utc_datetime(file_data.get('utc_datetime'), mandatory=True),
        hide=hide
    )


def import_file(
        file_data: FileData,
        object: Object,
        component: Component
) -> File:
    component_id = _get_or_create_component_id(file_data['component_uuid'])
    user_id = _get_or_create_user_id(file_data['user'])

    try:
        db_file = get_mutable_file(file_data['fed_id'], object.id, component_id)
        if db_file.user_id != user_id or db_file.data != file_data['data'] or db_file.utc_datetime != file_data['utc_datetime']:
            db_file.user_id = user_id
            db_file.data = file_data['data']
            db_file.utc_datetime = file_data['utc_datetime']
            db.session.commit()
            fed_logs.update_file(db_file.id, object.object_id, component.id)
    except errors.FileDoesNotExistError:
        assert component_id is not None
        db_file = create_fed_file(object.object_id, user_id, file_data['data'], None, file_data['utc_datetime'], file_data['fed_id'], component_id)
        fed_logs.import_file(db_file.id, db_file.object_id, component.id)
        if user_id is not None:
            object_log.upload_file(user_id=user_id, object_id=object.object_id, file_id=db_file.id, utc_datetime=file_data['utc_datetime'], is_imported=True)

    file = File.from_database(db_file)
    if file_data['hide'] is not None and not file.is_hidden:
        hide_user = _get_or_create_user_id(file_data['hide']['user'])
        hide_file(
            object_id=file.object_id,
            file_id=file.id,
            user_id=hide_user,
            reason=file_data['hide']['reason'],
            utc_datetime=file_data['hide']['utc_datetime']
        )
    return file


def parse_import_file(
        file_data: typing.Dict[str, typing.Any],
        object: Object,
        component: Component
) -> File:
    return import_file(parse_file(file_data), object, component)



# coding: utf-8
"""

"""
import json
import typing

import flask
import flask_login
from flask_wtf import FlaskForm
from wtforms.fields import StringField, BooleanField, SelectField, SelectMultipleField
from wtforms.validators import DataRequired, ValidationError
from flask_babel import _

from . import frontend
from .. import logic
from .utils import check_current_user_is_not_readonly
from ..utils import FlaskResponseT
from ..logic.utils import get_translated_text
from ..logic.components import get_component_or_none
from ..models import SciCatExportType


class ActionTypesSortingForm(FlaskForm):
    encoded_order = StringField("Order-String", [DataRequired()])

    def validate_encoded_order(form, field: StringField) -> None:
        valid_action_type_ids = [action_type.id for action_type in logic.action_types.get_action_types()]

        try:
            split_string = field.data.split(",")
            sorted_action_type_ids = list(map(int, split_string))
            for action_type_id in sorted_action_type_ids:
                if action_type_id not in valid_action_type_ids:
                    raise ValidationError(f"{action_type_id} is not a valid id")
        except ValueError:
            raise ValidationError("Invalid IDs.")

        field.data = sorted_action_type_ids


@frontend.route('/action_types/', methods=['GET', 'POST'])
@flask_login.login_required
def action_types() -> FlaskResponseT:
    if not flask_login.current_user.is_admin:
        return flask.abort(403)

    sorting_form = ActionTypesSortingForm()

    if sorting_form.validate_on_submit():
        check_current_user_is_not_readonly()

        logic.action_types.set_action_types_order(sorting_form.encoded_order.data)

    return flask.render_template(
        'action_types/action_types.html',
        action_types=logic.action_types.get_action_types(),
        sorting_form=sorting_form
    )


@frontend.route('/action_types/<int(signed=True):type_id>', methods=['GET', 'POST'])
@flask_login.login_required
def action_type(type_id: int) -> FlaskResponseT:
    if not flask_login.current_user.is_admin:
        return flask.abort(403)

    try:
        action_type = logic.action_types.get_action_type(
            action_type_id=type_id
        )
    except logic.errors.ActionTypeDoesNotExistError:
        return flask.abort(404)

    if flask.request.args.get('mode') == 'edit':
        if action_type.fed_id is not None:
            flask.flash(_('Editing imported action types is not yet supported.'), 'error')
            return flask.abort(403)
        else:
            return show_action_type_form(type_id)

    try:
        translations = logic.action_type_translations.get_action_type_translations_for_action_type(action_type.id)
    except logic.errors.ActionTypeTranslationDoesNotExistError:
        return flask.abort(404)

    return flask.render_template(
        'action_types/action_type.html',
        action_type=action_type,
        translations=translations,
        component=get_component_or_none(action_type.component_id)
    )


@frontend.route('/action_types/new', methods=['GET', 'POST'])
@flask_login.login_required
def new_action_type() -> FlaskResponseT:
    if not flask_login.current_user.is_admin:
        return flask.abort(403)
    return show_action_type_form(None)


class ActionTypeForm(FlaskForm):
    translations = StringField(validators=[DataRequired()])

    admin_only = BooleanField()
    show_on_frontpage = BooleanField()
    show_in_navbar = BooleanField()
    show_in_object_filters = BooleanField()
    enable_labels = BooleanField()
    enable_files = BooleanField()
    enable_locations = BooleanField()
    enable_publications = BooleanField()
    enable_comments = BooleanField()
    enable_activity_log = BooleanField()
    enable_related_objects = BooleanField()
    enable_project_link = BooleanField()
    enable_instrument_link = BooleanField()
    disable_create_objects = BooleanField()
    is_template = BooleanField()
    select_usable_in_action_types = SelectMultipleField(coerce=int)
    scicat_export_type = SelectField(choices=[
        ('none', '—'),
        (SciCatExportType.SAMPLE.name.lower(), _('Sample')),
        (SciCatExportType.RAW_DATASET.name.lower(), _('Raw Dataset')),
        (SciCatExportType.DERIVED_DATASET.name.lower(), _('Derived Dataset')),
    ])

    def set_select_usable_in_action_types_attributes(self) -> None:
        self.select_usable_in_action_types.choices = [
            (action_type.id, get_translated_text(action_type.name, default=_('Unnamed Action Type')))
            for action_type in logic.action_types.get_action_types()
            if not action_type.disable_create_objects
        ]

        self.select_usable_in_action_types.shared_action_types = [action_type.id
                                                                  for action_type in logic.action_types.get_action_types()
                                                                  if action_type.component_id is not None]


def show_action_type_form(type_id: typing.Optional[int]) -> FlaskResponseT:
    check_current_user_is_not_readonly()
    action_type_translations = []
    action_type_language_ids = []
    action_type_form = ActionTypeForm()
    action_type_form.set_select_usable_in_action_types_attributes()

    english = logic.languages.get_language(logic.languages.Language.ENGLISH)

    def validate_string(string: str) -> bool:
        try:
            return 0 < len(string) < 100
        except Exception:
            return False

    def validate_strings(strings: typing.Sequence[str]) -> bool:
        try:
            result = [validate_string(string) for string in strings]
            for temp in result:
                if temp is False:
                    return False
        except Exception:
            return False

        return True

    if type_id is not None:
        try:
            action_type = logic.action_types.get_action_type(type_id)
        except logic.errors.ActionTypeDoesNotExistError:
            return flask.abort(404)
        if 'action_submit' not in flask.request.form:
            action_type_translations = logic.action_type_translations.get_action_type_translations_for_action_type(action_type.id)
            action_type_language_ids = [translation.language_id for translation in action_type_translations]
            action_type_form.admin_only.data = action_type.admin_only
            action_type_form.show_on_frontpage.data = action_type.show_on_frontpage
            action_type_form.show_in_navbar.data = action_type.show_in_navbar
            action_type_form.show_in_object_filters.data = action_type.show_in_object_filters
            action_type_form.enable_labels.data = action_type.enable_labels
            action_type_form.enable_files.data = action_type.enable_files
            action_type_form.enable_locations.data = action_type.enable_locations
            action_type_form.enable_publications.data = action_type.enable_publications
            action_type_form.enable_comments.data = action_type.enable_comments
            action_type_form.enable_activity_log.data = action_type.enable_activity_log
            action_type_form.enable_related_objects.data = action_type.enable_related_objects
            action_type_form.enable_project_link.data = action_type.enable_project_link
            action_type_form.enable_instrument_link.data = action_type.enable_instrument_link
            action_type_form.disable_create_objects.data = action_type.disable_create_objects
            action_type_form.select_usable_in_action_types.data = [element.id for element in action_type.usable_in_action_types]
            action_type_form.is_template.data = action_type.is_template
            if action_type.scicat_export_type is None:
                action_type_form.scicat_export_type.data = 'none'
            else:
                action_type_form.scicat_export_type.data = action_type.scicat_export_type.name.lower()

    if action_type_form.validate_on_submit():
        if type_id is None:

            try:
                translation_data = json.loads(action_type_form.translations.data)
            except Exception:
                pass
            else:
                translation_keys = {'language_id', 'name', 'description',
                                    'object_name', 'object_name_plural', 'view_text', 'perform_text'}
                if not isinstance(translation_data, list):
                    translation_data = ()
                for translation in translation_data:
                    if not isinstance(translation, dict):
                        flask.flash(_('Please fill out the form.'), 'error')
                        return show_action_type_form(type_id)

                    if set(translation.keys()) != translation_keys:
                        flask.flash(_('Please fill out the form.'), 'error')
                        return show_action_type_form(type_id)

                    name = translation['name'].strip()
                    object_name = translation['object_name'].strip()
                    object_name_plural = translation['object_name_plural'].strip()
                    view_text = translation['view_text'].strip()
                    perform_text = translation['perform_text'].strip()

                    if not validate_strings([name, object_name, object_name_plural, view_text, perform_text]):
                        flask.flash(_('The fields for name, object name (singular and plural), view text and perform text are required.'))
                        return flask.render_template(
                            'action_types/action_type_form.html',
                            action_type_form=action_type_form,
                            action_type_translations=action_type_translations,
                            languages=logic.languages.get_languages(),
                            ENGLISH=english,
                            submit_text=_('Create') if type_id is None else _('Save')
                        )

                action_type = logic.action_types.create_action_type(
                    admin_only=action_type_form.admin_only.data,
                    show_on_frontpage=action_type_form.show_on_frontpage.data,
                    show_in_navbar=action_type_form.show_in_navbar.data,
                    show_in_object_filters=action_type_form.show_in_object_filters.data,
                    enable_labels=action_type_form.enable_labels.data,
                    enable_files=action_type_form.enable_files.data,
                    enable_locations=action_type_form.enable_locations.data,
                    enable_publications=action_type_form.enable_publications.data,
                    enable_comments=action_type_form.enable_comments.data,
                    enable_activity_log=action_type_form.enable_activity_log.data,
                    enable_related_objects=action_type_form.enable_related_objects.data,
                    enable_project_link=action_type_form.enable_project_link.data,
                    enable_instrument_link=action_type_form.enable_instrument_link.data,
                    disable_create_objects=action_type_form.disable_create_objects.data,
                    is_template=action_type_form.is_template.data,
                    usable_in_action_type_ids=action_type_form.select_usable_in_action_types.data,
                    scicat_export_type={
                        type.name.lower(): type
                        for type in [
                            SciCatExportType.SAMPLE,
                            SciCatExportType.RAW_DATASET,
                            SciCatExportType.DERIVED_DATASET
                        ]
                    }.get(action_type_form.scicat_export_type.data)
                )

                for translation in translation_data:
                    language_id = int(translation['language_id'])
                    name = translation['name'].strip()
                    description = translation['description'].strip()
                    object_name = translation['object_name'].strip()
                    object_name_plural = translation['object_name_plural'].strip()
                    view_text = translation['view_text'].strip()
                    perform_text = translation['perform_text'].strip()

                    logic.action_type_translations.set_action_type_translation(
                        action_type_id=action_type.id,
                        language_id=language_id,
                        name=name,
                        description=description,
                        object_name=object_name,
                        object_name_plural=object_name_plural,
                        view_text=view_text,
                        perform_text=perform_text
                    )
                # load action type with new translations
                action_type = logic.action_types.get_action_type(action_type.id)
                logic.action_types.add_action_type_to_order(action_type)
        else:

            action_type = logic.action_types.update_action_type(
                action_type_id=type_id,
                admin_only=action_type_form.admin_only.data,
                show_on_frontpage=action_type_form.show_on_frontpage.data,
                show_in_navbar=action_type_form.show_in_navbar.data,
                show_in_object_filters=action_type_form.show_in_object_filters.data,
                enable_labels=action_type_form.enable_labels.data,
                enable_files=action_type_form.enable_files.data,
                enable_locations=action_type_form.enable_locations.data,
                enable_publications=action_type_form.enable_publications.data,
                enable_comments=action_type_form.enable_comments.data,
                enable_activity_log=action_type_form.enable_activity_log.data,
                enable_related_objects=action_type_form.enable_related_objects.data,
                enable_project_link=action_type_form.enable_project_link.data,
                enable_instrument_link=action_type_form.enable_instrument_link.data,
                disable_create_objects=action_type_form.disable_create_objects.data,
                is_template=action_type_form.is_template.data,
                usable_in_action_type_ids=action_type_form.select_usable_in_action_types.data,
                scicat_export_type={
                    'sample': SciCatExportType.SAMPLE,
                    'raw_dataset': SciCatExportType.RAW_DATASET,
                    'derived_dataset': SciCatExportType.DERIVED_DATASET
                }.get(action_type_form.scicat_export_type.data)
            )

            try:
                translation_data = json.loads(action_type_form.translations.data)
            except Exception:
                pass
            else:
                translation_keys = {'language_id', 'name', 'description',
                                    'object_name', 'object_name_plural', 'view_text', 'perform_text'}
                if not isinstance(translation_data, list):
                    translation_data = ()

                valid_translations = set()
                for translation in translation_data:
                    if not isinstance(translation, dict):
                        continue
                    if set(translation.keys()) != translation_keys:
                        continue

                    language_id = int(translation['language_id'])
                    name = translation['name'].strip()
                    description = translation['description'].strip()
                    object_name = translation['object_name'].strip()
                    object_name_plural = translation['object_name_plural'].strip()
                    view_text = translation['view_text'].strip()
                    perform_text = translation['perform_text'].strip()

                    if not validate_strings([name, object_name, object_name_plural, view_text, perform_text]):
                        flask.flash(_('The fields for name, object name (singular and plural), view text and perform text are required.'))
                        return flask.render_template(
                            'action_types/action_type_form.html',
                            action_type_form=action_type_form,
                            action_type_translations=action_type_translations,
                            action_type_language_ids=action_type_language_ids,
                            languages=logic.languages.get_languages(),
                            ENGLISH=english,
                            submit_text=_('Create') if type_id is None else _('Save')
                        )

                    translation = logic.action_type_translations.set_action_type_translation(
                        action_type_id=action_type.id,
                        language_id=language_id,
                        name=name,
                        description=description,
                        object_name=object_name,
                        object_name_plural=object_name_plural,
                        view_text=view_text,
                        perform_text=perform_text)
                    valid_translations.add((translation.language_id, translation.action_type_id))
                # delete translations:
                for existing_translation in logic.action_type_translations.get_action_type_translations_for_action_type(action_type.id):
                    if (existing_translation.language_id, existing_translation.action_type_id) not in valid_translations:
                        logic.action_type_translations.delete_action_type_translation(
                            language_id=existing_translation.language_id,
                            action_type_id=existing_translation.action_type_id
                        )

        return flask.redirect(flask.url_for('.action_type', type_id=action_type.id))

    return flask.render_template(
        'action_types/action_type_form.html',
        action_type_form=action_type_form,
        action_type_translations=action_type_translations,
        action_type_language_ids=action_type_language_ids,
        languages=logic.languages.get_languages(),
        ENGLISH=english,
        submit_text=_('Create') if type_id is None else _('Save')
    )



# coding: utf-8
"""

"""

import datetime
import typing

from sqlalchemy.dialects import postgresql
from sqlalchemy.orm import relationship, Mapped, Query

from .. import db
from .objects import Objects
from .users import User
from .actions import ActionType
from .utils import Model

if typing.TYPE_CHECKING:
    from .components import Component
    from .topics import Topic


location_user_association_table = db.Table(
    'location_responsible_users',
    db.metadata,
    db.Column('location_id', db.Integer, db.ForeignKey('locations.id')),
    db.Column('user_id', db.Integer, db.ForeignKey('users.id'))
)

topic_location_association_table = db.Table(
    'location_topics',
    db.metadata,
    db.Column('location_id', db.Integer, db.ForeignKey('locations.id')),
    db.Column('topic_id', db.Integer, db.ForeignKey('topics.id'))
)


class LocationType(Model):
    __tablename__ = 'location_types'
    __table_args__ = (
        db.CheckConstraint(
            '(fed_id IS NOT NULL AND component_id IS NOT NULL) OR (name is NOT NULL)',
            name='location_types_not_null_check'
        ),
        db.UniqueConstraint('fed_id', 'component_id', name='location_types_fed_id_component_id_key'),
    )

    # default location type IDs
    # offset to -100 to allow later addition of new default location types
    # see: migrations/location_types_create_default_location_types.py
    LOCATION: int = -100 + 1

    id: Mapped[int] = db.Column(db.Integer, primary_key=True)
    name: Mapped[typing.Optional[typing.Dict[str, str]]] = db.Column(postgresql.JSON, nullable=True)
    location_name_singular: Mapped[typing.Optional[typing.Dict[str, str]]] = db.Column(postgresql.JSON, nullable=True)
    location_name_plural: Mapped[typing.Optional[typing.Dict[str, str]]] = db.Column(postgresql.JSON, nullable=True)
    admin_only: Mapped[bool] = db.Column(db.Boolean, nullable=False)
    enable_parent_location: Mapped[bool] = db.Column(db.Boolean, nullable=False)
    enable_sub_locations: Mapped[bool] = db.Column(db.Boolean, nullable=False)
    enable_object_assignments: Mapped[bool] = db.Column(db.Boolean, nullable=False)
    enable_responsible_users: Mapped[bool] = db.Column(db.Boolean, nullable=False)
    enable_instruments: Mapped[bool] = db.Column(db.Boolean, nullable=False)
    show_location_log: Mapped[bool] = db.Column(db.Boolean, nullable=False)
    fed_id: Mapped[typing.Optional[int]] = db.Column(db.Integer, nullable=True)
    component_id: Mapped[typing.Optional[int]] = db.Column(db.Integer, db.ForeignKey('components.id'), nullable=True)
    component: Mapped[typing.Optional['Component']] = relationship('Component')
    enable_capacities: Mapped[bool] = db.Column(db.Boolean, nullable=False)

    if typing.TYPE_CHECKING:
        query: typing.ClassVar[Query["LocationType"]]

    def __init__(
            self,
            name: typing.Optional[typing.Dict[str, str]],
            location_name_singular: typing.Optional[typing.Dict[str, str]],
            location_name_plural: typing.Optional[typing.Dict[str, str]],
            admin_only: bool,
            enable_parent_location: bool,
            enable_sub_locations: bool,
            enable_object_assignments: bool,
            enable_responsible_users: bool,
            enable_instruments: bool,
            enable_capacities: bool,
            show_location_log: bool,
            fed_id: typing.Optional[int] = None,
            component_id: typing.Optional[int] = None
    ) -> None:
        super().__init__(
            name=name,
            location_name_singular=location_name_singular,
            location_name_plural=location_name_plural,
            admin_only=admin_only,
            enable_parent_location=enable_parent_location,
            enable_sub_locations=enable_sub_locations,
            enable_object_assignments=enable_object_assignments,
            enable_responsible_users=enable_responsible_users,
            enable_instruments=enable_instruments,
            show_location_log=show_location_log,
            fed_id=fed_id,
            component_id=component_id,
            enable_capacities=enable_capacities
        )

    def __repr__(self) -> str:
        return f'<{type(self).__name__}(id={self.id}, name="{self.name}")>'


class Location(Model):
    __tablename__ = 'locations'
    __table_args__ = (
        db.CheckConstraint(
            '(fed_id IS NOT NULL AND component_id IS NOT NULL) OR (name is NOT NULL AND description IS NOT NULL)',
            name='locations_not_null_check'
        ),
        db.UniqueConstraint('fed_id', 'component_id', name='locations_fed_id_component_id_key'),
    )

    id: Mapped[int] = db.Column(db.Integer, primary_key=True)
    name: Mapped[typing.Optional[typing.Dict[str, str]]] = db.Column(postgresql.JSON, nullable=True)
    description: Mapped[typing.Optional[typing.Dict[str, str]]] = db.Column(postgresql.JSON, nullable=True)
    parent_location_id: Mapped[typing.Optional[int]] = db.Column(db.Integer, db.ForeignKey('locations.id'), nullable=True)
    fed_id: Mapped[typing.Optional[int]] = db.Column(db.Integer, nullable=True)
    component_id: Mapped[typing.Optional[int]] = db.Column(db.Integer, db.ForeignKey('components.id'), nullable=True)
    component: Mapped[typing.Optional['Component']] = relationship('Component')
    type_id: Mapped[int] = db.Column(db.Integer, db.ForeignKey('location_types.id'), nullable=False)
    type: Mapped[LocationType] = relationship('LocationType')
    responsible_users: Mapped[typing.List['User']] = relationship("User", secondary=location_user_association_table, order_by="User.name")
    is_hidden: Mapped[bool] = db.Column(db.Boolean, default=False, nullable=False)
    enable_object_assignments: Mapped[bool] = db.Column(db.Boolean, default=True, server_default=db.true(), nullable=False)
    topics: Mapped[typing.List['Topic']] = relationship('Topic', secondary=topic_location_association_table, back_populates='locations')

    if typing.TYPE_CHECKING:
        query: typing.ClassVar[Query["Location"]]

    def __init__(
            self,
            name: typing.Optional[typing.Dict[str, str]],
            description: typing.Optional[typing.Dict[str, str]],
            parent_location_id: typing.Optional[int] = None,
            fed_id: typing.Optional[int] = None,
            component_id: typing.Optional[int] = None,
            *,
            type_id: int
    ) -> None:
        super().__init__(
            name=name,
            description=description,
            parent_location_id=parent_location_id,
            fed_id=fed_id,
            component_id=component_id,
            type_id=type_id
        )

    def __repr__(self) -> str:
        return f'<{type(self).__name__}(id={self.id}, name="{self.name}", description="{self.description}", parent_location_id={self.parent_location_id})>'


class ObjectLocationAssignment(Model):
    __tablename__ = 'object_location_assignments'
    __table_args__ = (
        db.CheckConstraint(
            '(fed_id IS NOT NULL AND component_id IS NOT NULL) OR (description IS NOT NULL AND user_id IS NOT NULL AND utc_datetime IS NOT NULL)',
            name='object_location_assignments_not_null_check'
        ),
        db.CheckConstraint(
            'NOT (confirmed AND declined)',
            name='object_location_assignments_not_both_states_check'
        ),
        db.UniqueConstraint('fed_id', 'component_id', name='object_location_assignments_fed_id_component_id_key'),
    )

    id: Mapped[int] = db.Column(db.Integer, primary_key=True)
    object_id: Mapped[int] = db.Column(db.Integer, db.ForeignKey(Objects.object_id_column), nullable=False)
    location_id: Mapped[typing.Optional[int]] = db.Column(db.Integer, db.ForeignKey(Location.id), nullable=True)
    responsible_user_id: Mapped[typing.Optional[int]] = db.Column(db.Integer, db.ForeignKey(User.id), nullable=True)
    user_id: Mapped[typing.Optional[int]] = db.Column(db.Integer, db.ForeignKey(User.id), nullable=True)
    description: Mapped[typing.Optional[typing.Dict[str, str]]] = db.Column(postgresql.JSON, nullable=True)
    utc_datetime: Mapped[typing.Optional[datetime.datetime]] = db.Column(db.TIMESTAMP(timezone=True), nullable=True)
    confirmed: Mapped[bool] = db.Column(db.Boolean, nullable=False, default=False)
    location: Mapped[typing.Optional['Location']] = relationship('Location')
    fed_id: Mapped[typing.Optional[int]] = db.Column(db.Integer, nullable=True)
    component_id: Mapped[typing.Optional[int]] = db.Column(db.Integer, db.ForeignKey('components.id'), nullable=True)
    component: Mapped[typing.Optional['Component']] = relationship('Component')
    declined: Mapped[bool] = db.Column(db.Boolean, nullable=False, default=False)

    if typing.TYPE_CHECKING:
        query: typing.ClassVar[Query["ObjectLocationAssignment"]]

    def __init__(
            self,
            object_id: int,
            location_id: typing.Optional[int],
            user_id: typing.Optional[int],
            description: typing.Optional[typing.Dict[str, str]],
            utc_datetime: typing.Optional[datetime.datetime] = None,
            responsible_user_id: typing.Optional[int] = None,
            confirmed: bool = False,
            declined: bool = False,
            fed_id: typing.Optional[int] = None,
            component_id: typing.Optional[int] = None
    ) -> None:
        super().__init__(
            object_id=object_id,
            location_id=location_id,
            responsible_user_id=responsible_user_id,
            user_id=user_id,
            description=description,
            utc_datetime=utc_datetime if utc_datetime is not None else datetime.datetime.now(datetime.timezone.utc),
            confirmed=confirmed,
            fed_id=fed_id,
            component_id=component_id,
            declined=declined
        )

    def __repr__(self) -> str:
        return f'<{type(self).__name__}(id={self.id}, object_id={self.object_id}, location_id={self.location_id}, user_id={self.user_id}, responsible_user_id={self.responsible_user_id}, utc_datetime={self.utc_datetime}, description="{self.description}", confirmed={self.confirmed}, declined={self.declined})>'


class LocationCapacity(Model):
    __tablename__ = 'location_capacities'

    location_id: Mapped[int] = db.Column(db.Integer, db.ForeignKey(Location.id), primary_key=True)
    action_type_id: Mapped[int] = db.Column(db.Integer, db.ForeignKey(ActionType.id), primary_key=True)
    capacity: Mapped[typing.Optional[int]] = db.Column(db.Integer, nullable=True)

    if typing.TYPE_CHECKING:
        query: typing.ClassVar[Query["LocationCapacity"]]



# coding: utf-8
"""

"""

import flask
from flask_babel import _
from flask_wtf import FlaskForm
from wtforms import StringField, PasswordField, BooleanField, SubmitField, IntegerField
from wtforms.validators import InputRequired, Email, Length, EqualTo, DataRequired, ValidationError

from ..logic import authentication


class SigninForm(FlaskForm):
    username = StringField(validators=[InputRequired()])
    password = PasswordField(validators=[InputRequired()])
    remember_me = BooleanField()
    shared_device = BooleanField(default=False)


class SignoutForm(FlaskForm):
    pass


class InvitationForm(FlaskForm):
    email = StringField(validators=[InputRequired()])


class RequestPasswordResetForm(FlaskForm):
    email = StringField(validators=[InputRequired()])


class RegistrationForm(FlaskForm):
    email = StringField('Contact Email', validators=[
        InputRequired(message=_("Please enter your contact email.")),
        Email(message=_("Please enter your contact email."))
    ])
    name = StringField('Name', validators=[InputRequired()])
    password = PasswordField('New Password', validators=[
        InputRequired(),
        Length(min=3),
        EqualTo('password2', message=_('Passwords must match'))
    ])
    password2 = PasswordField('Confirm password', validators=[InputRequired()])
    submit = SubmitField('Register')

    def validate_name(self, field: StringField) -> None:
        if flask.current_app.config['ENFORCE_SPLIT_NAMES']:
            name = field.data
            if ', ' not in name[1:-1]:
                raise ValidationError(_("Please enter your name as: surname, given names."))

    def validate_password(self, field: PasswordField) -> None:
        if field.data and len(field.data.encode('utf-8')) > authentication.MAX_BCRYPT_PASSWORD_LENGTH:
            raise ValidationError(_('The password must be at most %(max_bcrypt_password_length)s bytes long.', max_bcrypt_password_length=authentication.MAX_BCRYPT_PASSWORD_LENGTH))


class PasswordForm(FlaskForm):
    password = PasswordField('New Password', validators=[
        InputRequired(),
        Length(min=3)
    ])
    submit = SubmitField('Change Password')


class AuthenticationPasswordForm(FlaskForm):
    id = IntegerField('Authentication_method_id', validators=[DataRequired()])
    password = PasswordField('New Password', validators=[
        InputRequired(),
        Length(min=3)
    ])
    password_confirmation = PasswordField(validators=[
        InputRequired(),
        Length(min=3),
        EqualTo('password')
    ])
    submit = SubmitField('Change Password')

    def validate_password(self, field: PasswordField) -> None:
        if field.data and len(field.data.encode('utf-8')) > authentication.MAX_BCRYPT_PASSWORD_LENGTH:
            raise ValidationError(_('The password must be at most %(max_bcrypt_password_length)s bytes long.', max_bcrypt_password_length=authentication.MAX_BCRYPT_PASSWORD_LENGTH))


class RevokeInvitationForm(FlaskForm):
    invitation_id = IntegerField(validators=[DataRequired()])



# coding: utf-8
"""
Drop default value for is_hidden column in locations table.
"""

import flask_sqlalchemy


def run(db: flask_sqlalchemy.SQLAlchemy) -> bool:
    # Perform migration
    db.session.execute(db.text("""
        ALTER TABLE locations
            ALTER COLUMN is_hidden DROP DEFAULT
    """))
    return True



import typing
import re
import string

from ..errors import ValidationError
from .utils import schema_iter


def _get_property_schema(
        root_schema: typing.Dict[str, typing.Any],
        absolute_path: typing.List[str]
) -> typing.Optional[typing.Dict[str, typing.Any]]:
    filter_property_path = tuple(
        path_element if isinstance(path_element, str) and path_element not in ('[?]', '*') and not re.match(r'^[-+]?[0-9]+$', path_element) else None
        for path_element in absolute_path
    )
    return dict(schema_iter(
        root_schema,
        filter_path_depth_limit=len(filter_property_path),
        filter_property_path=filter_property_path,
    )).get(filter_property_path)


def _simplify_absolute_path(
    absolute_path: typing.List[str]
) -> typing.Optional[typing.List[str]]:
    simplified_path: typing.List[str] = []
    for path_element in absolute_path:
        if path_element == '..':
            if not simplified_path:
                return None
            simplified_path.pop()
        else:
            simplified_path.append(path_element)
    return simplified_path


def validate_calculation(
        calculation: typing.Any,
        root_schema: typing.Dict[str, typing.Any],
        path: typing.List[str]
) -> None:
    if not isinstance(calculation, dict):
        raise ValidationError('calculation must be dict', path)
    valid_keys = {'property_names', 'formula', 'digits'}
    required_keys = {'property_names', 'formula'}
    calculation_keys = set(calculation.keys())
    invalid_keys = calculation_keys - valid_keys
    if invalid_keys:
        raise ValidationError(f'unexpected keys in calculation: {invalid_keys}', path)
    missing_keys = required_keys - calculation_keys
    if missing_keys:
        raise ValidationError(f'missing keys in calculation: {missing_keys}', path)
    own_schema = _get_property_schema(root_schema, path)
    if own_schema is None:
        raise ValidationError('internal schema validation error', path)
    if isinstance(own_schema.get('units'), list) and len(own_schema['units']) != 1:
        raise ValidationError('quantities with a calculation must have fixed units', path)
    if not isinstance(calculation['property_names'], dict):
        if not isinstance(calculation['property_names'], list):
            raise ValidationError('property_names must be dict or list', path)
        property_names = {
            property_name: property_name
            for property_name in calculation['property_names']
        }
    else:
        property_names = calculation['property_names']
    if not isinstance(calculation['formula'], str):
        raise ValidationError('formula must be string', path)
    if 'digits' in calculation and not isinstance(calculation['digits'], int):
        raise ValidationError('digits must be an integer value', path)
    if 'digits' in calculation and not 0 <= calculation['digits'] <= 15:
        raise ValidationError('digits must be an integer value between 0 and 15', path)
    for property_alias, relative_property_path in property_names.items():
        if isinstance(relative_property_path, str):
            relative_property_path = [relative_property_path]
        if not isinstance(property_alias, str) or '__' in property_alias or not all(c in (string.ascii_letters + string.digits + '_') for c in property_alias) or property_alias[0] not in string.ascii_letters or property_alias.endswith('_'):
            raise ValidationError('invalid property alias: ' + str(property_alias), path)
        if not all(type(path_element) in [str, int] for path_element in relative_property_path):
            raise ValidationError('expected list of strings and/or integers, but got: ' + str(relative_property_path), path)
        if '[?]' in relative_property_path:
            raise ValidationError('invalid property path: ' + str(relative_property_path), path)
        absolute_property_path = _simplify_absolute_path(path[:-1] + relative_property_path)
        if absolute_property_path is None:
            raise ValidationError('invalid property path: ' + str(relative_property_path), path)
        if path == absolute_property_path:
            # this will not cover all possible instances of self-referential
            # quantities due to array complexities, so those need to be
            # handled in javascript
            raise ValidationError('quantities with a calculation must not directly depend on themselves', path)
        property_schema = _get_property_schema(root_schema, absolute_property_path)
        if property_schema is None:
            raise ValidationError('invalid property path: ' + str(relative_property_path), path)
        if property_schema.get('type') != 'quantity':
            raise ValidationError('property_name does not belong to a quantity property', path)
        if isinstance(property_schema.get('units'), list) and len(property_schema['units']) != 1:
            raise ValidationError('quantities with a calculation must only depend on quantities with fixed units', path)



# coding: utf-8
"""
Settings logic.

This module implements the logic for reading and writing user settings using
the generic Settings model.
In addition to providing the basics for accessing arbitrary JSON-compatible
settings, this module also implements which actual settings exist and that
the user's settings contain the correct data types. These are defined using
the DEFAULT_SETTINGS global variable, which contains the default value for
each setting. If there is no default value for a setting, DEFAULT_SETTINGS
should contain None for it and code for data type verification needs to be
added to `_verify_setting`.
"""

import copy
import typing

from .. import db
from ..models import Settings
from . import users


DEFAULT_SETTINGS: typing.Dict[str, typing.Any] = {
    "OBJECTS_PER_PAGE": 25,
    "USE_SCHEMA_EDITOR": True,
    "SHOW_OBJECT_TYPE_AND_ID_ON_OBJECT_PAGE": None,
    "SHOW_OBJECT_TITLE": None,
    "FULL_WIDTH_OBJECTS_TABLE": None,
    "WORKFLOW_VIEW_MODALS": None,
    "WORKFLOW_VIEW_COLLAPSED": None,
    "USE_ADMIN_PERMISSIONS": False,
    "SHOW_INVITATION_LOG": False,
    "SHOW_HIDDEN_USERS_AS_ADMIN": True,
    "INSTRUMENT_LOG_ORDER_ASCENDING": True,
    "INSTRUMENT_LOG_ORDER_ATTRIBUTE": "datetime",
    "DATAVERSE_API_TOKEN": "",
    "SCICAT_API_TOKEN": "",
    "DEFAULT_INSTRUMENT_LIST_FILTERS": {},
    "DEFAULT_ACTION_LIST_FILTERS": {},
    "DEFAULT_OBJECT_LIST_FILTERS": {},
    "DEFAULT_OBJECT_LIST_OPTIONS": {},
    "DEFAULT_LOCATION_LIST_FILTERS": {},
    "AUTO_LC": True,
    "TIMEZONE": "UTC",
    "AUTO_TZ": True,
    "LOCALE": "en"
}


def get_user_settings(
        user_id: typing.Optional[int]
) -> typing.Dict[str, typing.Any]:
    """
    Get the settings for a user.

    This function will amend the user's settings with the default settings,
    so that code can rely on settings being available.

    :param user_id: the ID of an existing user, or None
    :return: the settings data
    :raise errors.UserDoesNotExistError: if the user does not exist
    """
    verified_data = copy.deepcopy(DEFAULT_SETTINGS)
    if user_id is None:
        return verified_data
    # ensure the user exists
    users.check_user_exists(user_id)
    settings = Settings.query.filter_by(user_id=user_id).first()
    if settings is not None:
        verified_data.update(_verify_settings(settings.data))
    return verified_data


def get_user_setting(
        user_id: typing.Optional[int],
        setting_name: str
) -> typing.Any:
    """
    Get a specific setting for a user.

    This function will return the default value for the setting, if the user's
    settings do not include it.

    :param user_id: the ID of an existing user, or None
    :param setting_name: the name of a setting
    :return: the settings data
    :raise errors.UserDoesNotExistError: if the user does not exist
    """
    default_value = copy.deepcopy(DEFAULT_SETTINGS.get(setting_name))
    if user_id is None:
        return default_value
    # ensure the user exists
    users.get_user(user_id)
    row: typing.Optional[typing.Tuple[typing.Any]] = db.session.query(Settings.data[setting_name]).filter(Settings.user_id == user_id).first()
    if row is not None and _verify_setting(setting_name, row[0]):
        return row[0]
    return default_value


def set_user_settings(user_id: int, data: typing.Dict[str, typing.Any]) -> None:
    """
    Set new settings for a user.

    This function will keep a user's previous settings if not overwritten.

    :param user_id: the ID of an existing user
    :param data: the settings data
    :raise errors.UserDoesNotExistError: if the user does not exist
    """
    # ensure the user exists
    users.check_user_exists(user_id)
    settings = Settings.query.filter_by(user_id=user_id).first()
    if settings is None:
        verified_data = {}
        settings = Settings(user_id, {})
    else:
        verified_data = _verify_settings(settings.data)
    verified_data.update(_verify_settings(data))
    settings.data = verified_data
    db.session.add(settings)
    db.session.commit()


def _verify_setting(key: str, value: typing.Any) -> bool:
    """
    Verify the value of an individual setting.

    :param key: the name of the setting
    :param value: the value of the setting
    :return: True, if the value is valid, False otherwise
    """
    if key not in DEFAULT_SETTINGS:
        return False
    default_value = DEFAULT_SETTINGS[key]
    if default_value is not None:
        if value is None:
            # settings that accept None must be listed here
            return key in {'OBJECTS_PER_PAGE'}
        return isinstance(value, type(default_value))
    # custom data type verification can be included here
    if key in {'SHOW_OBJECT_TITLE', 'SHOW_OBJECT_TYPE_AND_ID_ON_OBJECT_PAGE', 'FULL_WIDTH_OBJECTS_TABLE', 'WORKFLOW_VIEW_MODALS', 'WORKFLOW_VIEW_COLLAPSED'}:
        return value is None or isinstance(value, bool)
    return False


def _verify_settings(data: typing.Dict[str, typing.Any]) -> typing.Dict[str, typing.Any]:
    """
    Verify a dictionary of settings.

    :param data: the unverified settings data
    :return: the verified settings data, omitting any invalid settings
    """
    verified_data = {}
    for key, value in data.items():
        if _verify_setting(key, value):
            verified_data[key] = copy.deepcopy(value)
    return verified_data



# coding: utf-8
"""
Add missing NOT NULL constraints to columns in two_factor_authentication_methods table.
"""

import flask_sqlalchemy

from .utils import column_is_nullable


def run(db: flask_sqlalchemy.SQLAlchemy) -> bool:
    # Skip migration by condition
    if not column_is_nullable('two_factor_authentication_methods', 'user_id') and not column_is_nullable('two_factor_authentication_methods', 'data'):
        return False

    # Perform migration
    db.session.execute(db.text("""
        ALTER TABLE two_factor_authentication_methods
            ALTER COLUMN user_id SET NOT NULL,
            ALTER COLUMN data SET NOT NULL
    """))
    return True



# coding: utf-8
"""
Script for setting up demo data in a previously unused SampleDB installation.

Usage: sampledb set_up_demo
"""
import copy
import datetime
import io
import json
import os
import string
import sys
import random
import typing

from PIL import Image

import sampledb
from .. import create_app
from ..models import UserType, ActionType, Language
from ..logic.instruments import create_instrument, add_instrument_responsible_user
from ..logic.instrument_translations import set_instrument_translation
from ..logic.action_translations import set_action_translation
from ..logic.actions import create_action
from ..logic import groups, object_permissions, projects, comments, files


def main(arguments: typing.List[str]) -> None:
    if len(arguments) != 0:
        print(__doc__)
        sys.exit(1)
    app = create_app(include_dashboard=False)
    if not app.config.get("SERVER_NAME"):
        app.config["SERVER_NAME"] = "localhost:8000"
    with app.app_context():
        if sampledb.logic.actions.get_actions() or sampledb.logic.instruments.get_instruments() or len(sampledb.logic.users.get_users()) > 1:
            print("Error: database must be empty for demo", file=sys.stderr)
            sys.exit(1)

        data_directory = os.path.abspath(os.path.join(os.path.dirname(__file__), 'demo_data'))
        schema_directory = os.path.join(data_directory, 'schemas')
        objects_directory = os.path.join(data_directory, 'objects')

        if sampledb.logic.users.get_users():
            # admin might have been created from environment variables
            admin = sampledb.logic.users.get_users()[0]
        else:
            admin = sampledb.logic.users.create_user(name="Administrator", email="admin@example.com", type=UserType.PERSON)
            sampledb.logic.users.set_user_administrator(admin.id, True)

        instrument_responsible_user = sampledb.logic.users.create_user(name="Instrument Scientist", email="instrument@example.com", type=UserType.PERSON)
        basic_user = sampledb.logic.users.create_user(name="Basic User", email="basic@example.com", type=UserType.PERSON)
        api_user = sampledb.logic.users.create_user(name="API User", email="api@example.com", type=UserType.OTHER)
        sampledb.logic.authentication.add_other_authentication(api_user.id, 'api', 'password')

        # enable german for input
        german = sampledb.logic.languages.get_language(sampledb.models.Language.GERMAN)
        sampledb.logic.languages.update_language(
            language_id=german.id,
            names=german.names,
            lang_code=german.lang_code,
            datetime_format_datetime=german.datetime_format_datetime,
            datetime_format_moment=german.datetime_format_moment,
            datetime_format_moment_output=german.datetime_format_moment_output,
            date_format_moment_output=german.date_format_moment_output,
            enabled_for_input=True,
            enabled_for_user_interface=True
        )

        topic = sampledb.logic.topics.create_topic(
            name={
                "en": "Example Topic",
                "de": "Beispielthema"
            },
            description={
                "en": "This is an example topic.",
                "de": "Dies ist ein Beispielthema."
            },
            short_description={
                "en": "This is an example topic.",
                "de": "Dies ist ein Beispielthema."
            },
            show_on_frontpage=False,
            show_in_navbar=True,
            description_is_markdown=False,
            short_description_is_markdown=False
        )

        group_id = groups.create_group({"en": "Example Group", "de": "Beispielgruppe"},
                                       {"en": "This is an example group for testing purposes.", "de": "Dies ist eine Beispielgruppe für Testzwecke"},
                                       instrument_responsible_user.id).id

        project_id = projects.create_project({"en": "Example Project", "de": "Beispielprojekt"},
                                             {"en": "This is an example project",
                                              "de": "Dies ist ein Beispielprojekt"}, instrument_responsible_user.id).id

        project_id2 = projects.create_project({"en": "Example Project 2", "de": "Beispielprojekt 2"}, {"en": "This is another example project", "de": "Dies ist ein weiteres Beispielprojekt"},
                                              instrument_responsible_user.id).id
        projects.create_subproject_relationship(parent_project_id=project_id, child_project_id=project_id2,
                                                child_can_add_users_to_parent=True)

        projects.create_project({"en": "A-Example", "de": "A-Beispiel"}, {"en": "", "de": ""}, instrument_responsible_user.id)
        projects.create_project({"en": "C-Example", "de": "C-Beispiel"}, {"en": "", "de": ""}, instrument_responsible_user.id)
        project_id5 = projects.create_project({"en": "B-Example", "de": "B-Beispiel"}, {"en": "", "de": ""}, instrument_responsible_user.id).id
        project_id6 = projects.create_project({"en": "2-B-Example", "de": "2-B-Beispiel"}, {"en": "", "de": ""}, instrument_responsible_user.id).id
        project_id7 = projects.create_project({"en": "1-B-Example", "de": "1-B-Beispiel"}, {"en": "", "de": ""}, instrument_responsible_user.id).id
        projects.create_subproject_relationship(parent_project_id=project_id5, child_project_id=project_id6)
        projects.create_subproject_relationship(parent_project_id=project_id5, child_project_id=project_id7)
        projects.add_user_to_project(project_id5, admin.id, sampledb.models.Permissions.READ)

        markdown_notes = """# Header
This example shows how Markdown can be used for instrument Notes.

## Subheader

*italics* **bold**


| A | B | C |
|--:|:-:|---|
| Example | 100˚C | 5µm |
| Data | 110˚C | 6µm |
            """
        instrument = create_instrument(
            users_can_create_log_entries=True,
            notes_is_markdown=True
        )
        sampledb.logic.topics.set_instrument_topics(instrument.id, [topic.id])
        set_instrument_translation(
            language_id=sampledb.models.Language.ENGLISH,
            instrument_id=instrument.id,
            name="OMBE I",
            description="This is an example instrument",
            short_description="This is the short description",
            notes=markdown_notes,
        )
        set_instrument_translation(
            language_id=sampledb.models.Language.GERMAN,
            instrument_id=instrument.id,
            name="OMBE I",
            description="Dies ist ein Beispiel Instrument",
            short_description="Dies ist die kurze Beschreibung",
            notes=markdown_notes,
        )
        add_instrument_responsible_user(instrument.id, instrument_responsible_user.id)
        log_category_error = sampledb.logic.instrument_log_entries.create_instrument_log_category(
            instrument_id=instrument.id,
            title='Error',
            theme=sampledb.models.instrument_log_entries.InstrumentLogCategoryTheme.RED
        )
        log_category_warning = sampledb.logic.instrument_log_entries.create_instrument_log_category(
            instrument_id=instrument.id,
            title='Warning',
            theme=sampledb.models.instrument_log_entries.InstrumentLogCategoryTheme.YELLOW
        )
        log_category_success = sampledb.logic.instrument_log_entries.create_instrument_log_category(
            instrument_id=instrument.id,
            title='Success',
            theme=sampledb.models.instrument_log_entries.InstrumentLogCategoryTheme.GREEN
        )
        log_category_other = sampledb.logic.instrument_log_entries.create_instrument_log_category(
            instrument_id=instrument.id,
            title='Other',
            theme=sampledb.models.instrument_log_entries.InstrumentLogCategoryTheme.BLUE
        )
        log_category_normal = sampledb.logic.instrument_log_entries.create_instrument_log_category(
            instrument_id=instrument.id,
            title='Normal',
            theme=sampledb.models.instrument_log_entries.InstrumentLogCategoryTheme.GRAY
        )
        for category in sampledb.logic.instrument_log_entries.get_instrument_log_categories(instrument.id):
            sampledb.logic.instrument_log_entries.create_instrument_log_entry(
                instrument.id, basic_user.id, "This is an example instrument log entry",
                [category.id]
            )
        sampledb.logic.instrument_log_entries.create_instrument_log_entry(
            instrument.id, basic_user.id, "This is an example instrument log entry"
        )
        sampledb.logic.instrument_log_entries.create_instrument_log_entry(
            instrument.id, basic_user.id, "This is an example instrument log entry",
            [log_category_other.id]
        )
        log_entry = sampledb.logic.instrument_log_entries.create_instrument_log_entry(
            instrument.id, basic_user.id, "This is an example instrument log entry",
            [log_category_error.id, log_category_warning.id, log_category_normal.id, log_category_success.id]
        )
        with open(os.path.join(os.path.dirname(sampledb.__file__), 'static/sampledb/img/ghs01.png'), 'rb') as ghs01_file:
            ghs01_image = ghs01_file.read()
        sampledb.logic.instrument_log_entries.create_instrument_log_file_attachment(
            instrument_log_entry_id=log_entry.id,
            file_name="ghs01.png",
            content=ghs01_image
        )
        with open(os.path.join(os.path.dirname(sampledb.__file__), 'static/sampledb/img/ghs02.png'), 'rb') as ghs02_file:
            ghs02_image = ghs02_file.read()
        sampledb.logic.instrument_log_entries.create_instrument_log_file_attachment(
            instrument_log_entry_id=log_entry.id,
            file_name="ghs02.png",
            content=ghs02_image
        )
        sampledb.logic.instrument_log_entries.create_instrument_log_file_attachment(
            instrument_log_entry_id=log_entry.id,
            file_name="test.txt",
            content="This is a test".encode('utf-8')
        )

        with open(os.path.join(schema_directory, 'ombe_measurement.sampledb.json'), 'r', encoding='utf-8') as schema_file:
            schema = json.load(schema_file)
        instrument_action = create_action(
            action_type_id=ActionType.SAMPLE_CREATION,
            schema=schema,
            instrument_id=instrument.id
        )
        sampledb.logic.topics.set_action_topics(instrument_action.id, [topic.id])

        set_action_translation(Language.ENGLISH, instrument_action.id, "Sample Creation", "This is an example action")

        sampledb.logic.action_permissions.set_action_permissions_for_all_users(instrument_action.id, sampledb.models.Permissions.READ)
        independent_action = create_action(
            action_type_id=ActionType.SAMPLE_CREATION,
            schema=schema
        )
        set_action_translation(Language.ENGLISH, independent_action.id, "Alternative Process", "This is an example action")

        sampledb.logic.action_permissions.set_action_permissions_for_all_users(independent_action.id, sampledb.models.Permissions.READ)
        with open(os.path.join(schema_directory, 'ombe_measurement_batch.sampledb.json'), 'r', encoding='utf-8') as schema_file:
            batch_schema = json.load(schema_file)
        action = create_action(
            action_type_id=ActionType.SAMPLE_CREATION,
            schema=batch_schema,
            instrument_id=instrument.id
        )
        set_action_translation(Language.ENGLISH, action.id, "Sample Creation (Batch)", "This is an example action")

        sampledb.logic.action_permissions.set_action_permissions_for_all_users(action.id, sampledb.models.Permissions.READ)
        sampledb.db.session.commit()

        with open(os.path.join(objects_directory, 'ombe-1.sampledb.json'), 'r', encoding='utf-8') as data_file:
            data = json.load(data_file)
        instrument_object = sampledb.logic.objects.create_object(data=data, schema=schema, user_id=instrument_responsible_user.id, action_id=instrument_action.id)
        independent_object = sampledb.logic.objects.create_object(data=data, schema=schema, user_id=instrument_responsible_user.id, action_id=independent_action.id)
        comments.create_comment(instrument_object.id, instrument_responsible_user.id, 'This comment is very long. ' * 20 + '\n' + 'This comment has three paragraphs. ' * 20 + '\n' + '\n' + 'This comment has three paragraphs. ' * 20)
        comments.create_comment(instrument_object.id, instrument_responsible_user.id, 'This is another, shorter comment')
        files.create_database_file(instrument_object.id, instrument_responsible_user.id, 'example.txt', lambda stream: typing.cast(None, stream.write("Dies ist ein Test".encode('utf-8'))))
        files.create_database_file(instrument_object.id, instrument_responsible_user.id, 'demo.png', lambda stream: typing.cast(None, stream.write(ghs01_image)))
        files.update_file_information(instrument_object.id, 1, instrument_responsible_user.id, 'Example File', 'This is a file description.')
        files.create_url_file(instrument_object.id, instrument_responsible_user.id, 'http://iffsamples.fz-juelich.de/')
        sampledb.logic.publications.link_publication_to_object(instrument_responsible_user.id, instrument_object.id, '10.5281/zenodo.4012175', 'sciapp/sampledb', 'Example')

        projects.link_project_and_object(project_id, instrument_object.object_id, instrument_responsible_user.id)

        object_permissions.set_object_permissions_for_anonymous_users(instrument_object.id, sampledb.models.Permissions.READ)

        with open(os.path.join(schema_directory, 'ombe_measurement_updated.sampledb.json'), 'r', encoding='utf-8') as schema_file:
            schema = json.load(schema_file)
        sampledb.logic.actions.update_action(
            action_id=instrument_action.id,
            schema=schema
        )
        sampledb.logic.action_translations.set_action_translation(Language.ENGLISH, instrument_action.id, "Updated Sample Creation", "", "")

        object_permissions.set_group_object_permissions(independent_object.object_id, group_id, sampledb.models.Permissions.READ)

        instrument = create_instrument()
        set_instrument_translation(Language.ENGLISH, instrument.id, name="XRR", description="X-Ray Reflectometry")

        add_instrument_responsible_user(instrument.id, instrument_responsible_user.id)
        with open(os.path.join(schema_directory, 'xrr_measurement.sampledb.json'), 'r', encoding='utf-8') as schema_file:
            schema = json.load(schema_file)
        instrument_action = create_action(
            action_type_id=ActionType.MEASUREMENT,
            schema=schema,
            instrument_id=instrument.id
        )
        set_action_translation(Language.ENGLISH, instrument_action.id, "XRR Measurement", "", )

        sampledb.logic.action_permissions.set_action_permissions_for_all_users(instrument_action.id, sampledb.models.Permissions.READ)
        with open(os.path.join(schema_directory, 'searchable_quantity.json'), 'r', encoding='utf-8') as schema_file:
            schema = json.load(schema_file)
        action = create_action(
            action_type_id=ActionType.SAMPLE_CREATION,
            schema=schema
        )
        set_action_translation(Language.ENGLISH, action.id, "Searchable Object", "")

        sampledb.logic.action_permissions.set_action_permissions_for_all_users(action.id, sampledb.models.Permissions.READ)
        independent_object = sampledb.logic.objects.create_object(data={
            "name": {
                "_type": "text",
                "text": {"en": "TEST-1", "de": "TEST-1"}
            },
            "tags": {
                "_type": "tags",
                "tags": ["tag1", "tag2"]
            },
            "mass": {
                "_type": "quantity",
                "dimensionality": "[mass]",
                "magnitude_in_base_units": 0.00001,
                "units": "mg"
            }
        }, schema=schema, user_id=instrument_responsible_user.id, action_id=action.id)
        object_permissions.set_group_object_permissions(independent_object.object_id, group_id, sampledb.models.Permissions.READ)
        object_permissions.set_user_object_permissions(independent_object.object_id, api_user.id, sampledb.models.Permissions.WRITE)
        independent_object = sampledb.logic.objects.create_object(data={
            "name": {
                "_type": "text",
                "text": {'en': "TEST-2"}
            },
            "tags": {
                "_type": "tags",
                "tags": ["tag2", "tag3"]
            },
            "mass": {
                "_type": "quantity",
                "dimensionality": "[mass]",
                "magnitude_in_base_units": 0.000005,
                "units": "mg"
            }
        }, schema=schema, user_id=instrument_responsible_user.id, action_id=action.id)
        object_permissions.set_group_object_permissions(independent_object.object_id, group_id, sampledb.models.Permissions.READ)
        sampledb.db.session.commit()

        instrument = create_instrument()
        set_instrument_translation(Language.ENGLISH, instrument.id, name="MPMS SQUID", description="MPMS SQUID Magnetometer JCNS-2")

        add_instrument_responsible_user(instrument.id, instrument_responsible_user.id)
        with open(os.path.join(schema_directory, 'squid_measurement.sampledb.json'), 'r', encoding='utf-8') as schema_file:
            schema = json.load(schema_file)
        instrument_action = create_action(
            action_type_id=ActionType.MEASUREMENT,
            schema=schema,
            instrument_id=instrument.id
        )
        set_action_translation(Language.ENGLISH, instrument_action.id, "Perform Measurement", "")
        sampledb.logic.action_permissions.set_action_permissions_for_all_users(instrument_action.id, sampledb.models.Permissions.READ)
        sampledb.db.session.commit()

        instrument = create_instrument()
        set_instrument_translation(Language.ENGLISH, instrument.id, name="Powder Diffractometer", description="Huber Imaging Plate Guinier Camera G670 at JCNS-2")
        add_instrument_responsible_user(instrument.id, instrument_responsible_user.id)
        with open(os.path.join(schema_directory, 'powder_diffractometer_measurement.sampledb.json'), 'r', encoding='utf-8') as schema_file:
            schema = json.load(schema_file)
        instrument_action = create_action(
            action_type_id=ActionType.MEASUREMENT,
            schema=schema,
            instrument_id=instrument.id
        )
        set_action_translation(Language.ENGLISH, instrument_action.id, "Perform Measurement", "")
        sampledb.logic.action_permissions.set_action_permissions_for_all_users(instrument_action.id, sampledb.models.Permissions.READ)
        sampledb.db.session.commit()

        instrument = create_instrument()
        set_instrument_translation(Language.ENGLISH, instrument.id, name="GALAXI", description="Gallium Anode Low-Angle X-ray Instrument")
        add_instrument_responsible_user(instrument.id, instrument_responsible_user.id)
        with open(os.path.join(schema_directory, 'galaxi_measurement.sampledb.json'), 'r', encoding='utf-8') as schema_file:
            schema = json.load(schema_file)
        instrument_action = create_action(
            action_type_id=ActionType.MEASUREMENT,
            schema=schema,
            instrument_id=instrument.id
        )
        set_action_translation(Language.ENGLISH, instrument_action.id, "Perform Measurement", "")
        sampledb.logic.action_permissions.set_action_permissions_for_all_users(instrument_action.id, sampledb.models.Permissions.READ)
        sampledb.db.session.commit()

        with open(os.path.join(schema_directory, 'calculation.sampledb.json'), 'r', encoding='utf-8') as schema_file:
            schema = json.load(schema_file)
        action = create_action(
            action_type_id=ActionType.SAMPLE_CREATION,
            schema=schema
        )
        set_action_translation(Language.ENGLISH, action.id, "Calculation Demo Action", "", )
        sampledb.logic.action_permissions.set_action_permissions_for_all_users(action.id, sampledb.models.Permissions.READ)
        sampledb.db.session.commit()

        with open(os.path.join(schema_directory, 'other_sample.sampledb.json'), 'r', encoding='utf-8') as schema_file:
            schema = json.load(schema_file)
        action = create_action(
            action_type_id=ActionType.SAMPLE_CREATION,
            schema=schema
        )
        set_action_translation(Language.ENGLISH, action.id, "Other Sample", "", )
        sampledb.logic.action_permissions.set_action_permissions_for_all_users(action.id, sampledb.models.Permissions.READ)
        sampledb.db.session.commit()

        sample_schema: typing.Dict[str, typing.Any] = {
            'title': 'Sample Information',
            'type': 'object',
            'properties': {
                'name': {
                    'title': {'en': 'Object Name', 'de': 'Objektname'},
                    'type': 'text',
                    'languages': 'all'
                },
                'sample': {
                    'title': {'en': 'Sample', 'de': 'Probe'},
                    'type': 'sample'
                }
            },
            'required': ['name']
        }
        sample_action = sampledb.logic.actions.create_action(
            action_type_id=ActionType.SAMPLE_CREATION,
            schema=sample_schema
        )
        sample_schema['workflow_views'] = [
            {
                'title': {'en': 'Referencing Measurements (sorted by property "datetime")'},
                'referencing_action_type_id': -98,
                'referenced_action_id': [],
                'sorting_properties': ['datetime']
            },
            {
                'title': {'en': 'Referencing Measurements (recursion)'},
                'referencing_action_type_id': [ActionType.MEASUREMENT, ActionType.SAMPLE_CREATION],
                'referenced_action_id': [],
                'sorting_properties': ['datetime'],
                'recursion_filters': {
                    'referenced_action_id': sample_action.id,
                    'referenced_action_type_id': ActionType.MEASUREMENT,
                    'referenced_filter_operator': 'or',
                    'referencing_action_id': sample_action.id,
                    'referencing_action_type_id': ActionType.MEASUREMENT,
                    'referencing_filter_operator': 'and',
                    'max_depth': 2,
                }
            },
            {
                'title': {'en': f'Referenced Samples from Action #{sample_action.id}'},
                'referenced_action_id': sample_action.id,
                'referencing_action_id': [],
                'show_action_info': False,
            },
            {
                'title': {'en': 'All Related Objects'}
            }
        ]
        sampledb.logic.actions.update_action(action_id=sample_action.id, schema=sample_schema)
        set_action_translation(Language.ENGLISH, sample_action.id, name="sample_action", description="")
        sampledb.logic.action_permissions.set_action_permissions_for_all_users(sample_action.id, sampledb.models.Permissions.READ)
        measurement_action = sampledb.logic.actions.create_action(
            action_type_id=ActionType.MEASUREMENT,
            schema={
                'title': 'Measurement Information',
                'type': 'object',
                'properties': {
                    'name': {
                        'title': 'Object Name',
                        'type': 'text',
                        'languages': ['en', 'de']
                    },
                    'datetime': {
                        'title': 'Measurement Date/Time',
                        'type': 'datetime'
                    },
                    'sample': {
                        'title': 'Sample',
                        'type': 'sample',
                        'style': 'include'
                    },
                    'comment': {
                        'title': {'en': 'Comment', 'de': 'Kommentar'},
                        'type': 'text',
                        'markdown': True,
                        'languages': "all"
                    },
                    "tags": {
                        "title": {"en": "Tags", "de": "Tags"},
                        "type": "tags"
                    }
                },
                'required': ['name', 'tags'],
                'workflow_show_more': ['sample', 'comment']
            }
        )
        set_action_translation(Language.ENGLISH, measurement_action.id, name="measurement_action", description="")
        sampledb.logic.action_permissions.set_action_permissions_for_all_users(measurement_action.id, sampledb.models.Permissions.READ)
        data = {
            'name': {
                '_type': 'text',
                'text': {'en': 'Object 1', 'de': 'Objekt 1'}
            }
        }
        object = sampledb.logic.objects.create_object(sample_action.id, data, basic_user.id)
        sampledb.logic.object_permissions.set_object_permissions_for_all_users(object.id, sampledb.models.Permissions.READ)
        data = {
            'name': {
                '_type': 'text',
                'text': {'en': 'Object 2', 'de': 'Objekt 2'}
            },
            'sample': {
                '_type': 'sample',
                'object_id': object.id
            }
        }
        sample = sampledb.logic.objects.create_object(sample_action.id, data, basic_user.id)
        sampledb.logic.object_permissions.set_object_permissions_for_all_users(sample.id, sampledb.models.Permissions.READ)
        data = {
            'name': {
                '_type': 'text',
                'text': {'en': 'Object 1', 'de': 'Objekt 1'}
            },
            'sample': {
                '_type': 'sample',
                'object_id': sample.id
            }
        }
        sampledb.logic.objects.update_object(object.id, data, basic_user.id)
        data = {
            'name': {
                '_type': 'text',
                'text': {"en": 'Measurement', 'de': 'Messung'}
            },
            'datetime': {
                '_type': 'datetime',
                'utc_datetime': datetime.datetime.now(tz=datetime.timezone.utc).strftime('%Y-%m-%d %H:%M:%S')
            },
            'sample': {
                '_type': 'sample',
                'object_id': instrument_object.id
            },
            'comment': {
                '_type': 'text',
                'text': {'en': 'This is a test.\nThis **is** a *second* line.\n\nThis line follows an empty line.'},
                'is_markdown': True
            },
            'tags': {
                '_type': 'tags',
                'tags': ['example_tag', 'other_tag', 'tag3']
            }
        }
        measurement = sampledb.logic.objects.create_object(measurement_action.id, data, basic_user.id)
        sampledb.logic.object_permissions.set_object_permissions_for_all_users(measurement.id, sampledb.models.Permissions.READ)
        data = {
            'name': {
                '_type': 'text',
                'text': {'en': 'Measurement 2', 'de': 'Messung 2'}
            },
            'datetime': {
                '_type': 'datetime',
                'utc_datetime': datetime.datetime.now(tz=datetime.timezone.utc).strftime('%Y-%m-%d %H:%M:%S')
            },
            'sample': {
                '_type': 'sample',
                'object_id': sample.id
            },
            'comment': {
                '_type': 'text',
                'text': {'en': 'This is a test.\nThis is a second line.\n\nThis line follows an empty line.'}
            },
            'tags': {
                '_type': 'tags',
                'tags': []
            }
        }
        measurement = sampledb.logic.objects.create_object(measurement_action.id, data, instrument_responsible_user.id)
        sampledb.logic.object_permissions.set_object_permissions_for_all_users(measurement.id, sampledb.models.Permissions.READ)
        data = {
            'name': {
                '_type': 'text',
                'text': {'en': 'Measurement 3', 'de': 'Messung 3'}
            },
            'datetime': {
                '_type': 'datetime',
                'utc_datetime': (datetime.datetime.now(tz=datetime.timezone.utc) - datetime.timedelta(hours=1)).strftime('%Y-%m-%d %H:%M:%S')
            },
            'sample': {
                '_type': 'sample',
                'object_id': sample.id
            },
            'comment': {
                '_type': 'text',
                'text': {'en': 'This is a test.\nThis is a second line.\n\nThis line follows an empty line.'}
            },
            'tags': {
                '_type': 'tags',
                'tags': []
            }
        }
        measurement = sampledb.logic.objects.create_object(measurement_action.id, data, instrument_responsible_user.id)
        sampledb.logic.object_permissions.set_object_permissions_for_all_users(measurement.id, sampledb.models.Permissions.READ)

        with open(os.path.join(schema_directory, 'plotly.json'), 'r', encoding='utf-8') as schema_file:
            schema = json.load(schema_file)
        plotly_action = create_action(
            action_type_id=ActionType.SAMPLE_CREATION,
            schema=schema,
            instrument_id=instrument.id
        )
        set_action_translation(Language.ENGLISH, plotly_action.id, name="Plotly Example Action", description="")
        sampledb.logic.action_permissions.set_action_permissions_for_all_users(plotly_action.id, sampledb.models.Permissions.READ)

        with open(os.path.join(objects_directory, 'plotly-example-data1.sampledb.json'), 'r', encoding='utf-8') as data_file:
            example_data = json.load(data_file)

        data = {
            "name": {
                "_type": "text",
                "text": "Plotly Example Data #1"
            },
            "plot1": {
                "_type": "plotly_chart",
                "plotly": example_data
            }
        }
        plotly_object = sampledb.logic.objects.create_object(plotly_action.id, data, basic_user.id)
        sampledb.logic.object_permissions.set_object_permissions_for_all_users(plotly_object.id, sampledb.models.Permissions.READ)

        with open(os.path.join(schema_directory, 'plotly_array.json'), 'r', encoding='utf-8') as schema_file:
            schema = json.load(schema_file)
        plotly_array_action = create_action(
            action_type_id=ActionType.SAMPLE_CREATION,
            schema=schema
        )
        set_action_translation(Language.ENGLISH, plotly_array_action.id, name="Plotly Array Example Action", description="")
        sampledb.logic.action_permissions.set_action_permissions_for_all_users(plotly_array_action.id, sampledb.models.Permissions.READ)

        with open(os.path.join(objects_directory, 'plotly-example-data2.sampledb.json'), 'r', encoding='utf-8') as data_file:
            example_data2 = json.load(data_file)
        with open(os.path.join(objects_directory, 'plotly-example-data3.sampledb.json'), 'r', encoding='utf-8') as data_file:
            example_data3 = json.load(data_file)

        data = {
            "name": {
                "_type": "text",
                "text": "Plotly Array Example"
            },
            "plotlist": [
                {
                    "_type": "plotly_chart",
                    "plotly": example_data
                },
                {
                    "_type": "plotly_chart",
                    "plotly": example_data2
                },
                {
                    "_type": "plotly_chart",
                    "plotly": example_data3
                }
            ]
        }
        plotly_object = sampledb.logic.objects.create_object(plotly_array_action.id, data, basic_user.id)
        sampledb.logic.object_permissions.set_object_permissions_for_all_users(plotly_object.id, sampledb.models.Permissions.READ)
        sampledb.db.session.commit()

        campus = sampledb.logic.locations.create_location({"en": "Campus"}, {"en": "Max Mustermann Campus"}, None, instrument_responsible_user.id, type_id=sampledb.models.LocationType.LOCATION)
        building_a = sampledb.logic.locations.create_location({"en": "Building A", "de": "Gebäude A"}, {"en": "Building A on Max Mustermann Campus", "de": "Gebäude A auf dem Max Mustermann Campus"}, campus.id, instrument_responsible_user.id, type_id=sampledb.models.LocationType.LOCATION)
        room_42a = sampledb.logic.locations.create_location({"en": "Room 42a", "de": "Raum 42a"}, {"en": "Building A, Room 42a", "de": "Gebäude A, Raum 42a"}, building_a.id, instrument_responsible_user.id, type_id=sampledb.models.LocationType.LOCATION)
        room_42b = sampledb.logic.locations.create_location({"en": "Room 42b", "de": "Raum 42b"}, {"en": "Building A, Room 42b", "de": "Gebäude A, Raum 42b"}, building_a.id, instrument_responsible_user.id, type_id=sampledb.models.LocationType.LOCATION)
        sampledb.logic.locations.assign_location_to_object(measurement.id, room_42a.id, None, instrument_responsible_user.id, {"en": "Temporarily stored on table\n\nSome other text", "de": "Temporär auf einem Tisch gelagert \n\n Irgendein anderer Text"})
        sampledb.logic.locations.assign_location_to_object(measurement.id, room_42b.id, instrument_responsible_user.id, basic_user.id, {"en": "Stored in shelf K", "de": "In Regal K gelagert"})
        sampledb.logic.notifications.create_other_notification(instrument_responsible_user.id, "This is a demo.")
        sampledb.logic.object_permissions.set_user_object_permissions(independent_object.id, instrument_responsible_user.id, sampledb.models.Permissions.GRANT)
        sampledb.logic.notifications.create_notification_for_having_received_an_objects_permissions_request(instrument_responsible_user.id, independent_object.id, admin.id)
        sampledb.logic.location_permissions.set_location_permissions_for_all_users(campus.id, sampledb.models.Permissions.WRITE)
        sampledb.logic.location_permissions.set_location_permissions_for_all_users(building_a.id, sampledb.models.Permissions.WRITE)
        sampledb.logic.location_permissions.set_location_permissions_for_all_users(room_42a.id, sampledb.models.Permissions.WRITE)
        sampledb.logic.location_permissions.set_location_permissions_for_all_users(room_42b.id, sampledb.models.Permissions.WRITE)
        sampledb.logic.instruments.set_instrument_location(instrument.id, room_42a.id)
        sampledb.logic.topics.set_location_topics(room_42a.id, [topic.id])

        container_type = sampledb.logic.locations.create_location_type(
            name={"en": "Container"},
            location_name_singular={"en": "Container"},
            location_name_plural={"en": "Containers"},
            admin_only=True,
            enable_parent_location=True,
            enable_sub_locations=False,
            enable_object_assignments=True,
            enable_responsible_users=True,
            enable_instruments=False,
            enable_capacities=True,
            show_location_log=True,
        )
        container = sampledb.logic.locations.create_location(
            name={"en": "Box 1"},
            description={"en": "An example container"},
            parent_location_id=room_42b.id,
            user_id=instrument_responsible_user.id,
            type_id=container_type.id
        )
        sampledb.logic.location_permissions.set_location_permissions_for_all_users(container.id, sampledb.models.Permissions.WRITE)
        sampledb.logic.locations.set_location_responsible_users(container.id, [instrument_responsible_user.id, admin.id])
        sampledb.logic.locations.set_location_capacity(container.id, sampledb.models.ActionType.SAMPLE_CREATION, 2)

        for object in sampledb.logic.objects.get_objects():
            sampledb.logic.instrument_log_entries.create_instrument_log_object_attachment(
                instrument_log_entry_id=log_entry.id,
                object_id=object.id
            )
        action = sampledb.logic.actions.create_action(
            action_type_id=ActionType.SAMPLE_CREATION,
            schema={
                'title': 'Example Object',
                'type': 'object',
                'properties': {
                    'name': {
                        'title': 'Object Name',
                        'type': 'text',
                        'languages': ['en', 'de']
                    },
                    'dropdown': {
                        'title': {'en': 'English Title', 'de': 'Deutscher Titel'},
                        'type': 'text',
                        'choices': [
                            {'en': 'en 1', 'de': 'de 1'},
                            {'en': 'en 2', 'de': 'de 2'},
                            {'en': 'en 3'}
                        ],
                        'default': {'en': 'en 2', 'de': 'de 2'},
                        'note': 'Select option 1.'
                    },
                    'user': {
                        'title': {'en': 'User', 'de': 'Nutzer'},
                        'type': 'user',
                        'note': 'Do not select a user.'
                    },
                    'checkbox': {
                        'title': {'en': 'Checkbox', 'de': 'Checkbox'},
                        'type': 'bool',
                        'note': 'Check this checkbox.'
                    },
                    'object': {
                        'title': {'en': 'Object', 'de': 'Objekt'},
                        'type': 'object_reference',
                        'note': 'Select object #1.',
                        'action_type_id': [-99, -98],
                        'action_id': [1, 3]
                    },
                    'conditional_text': {
                        'title': 'Conditional Name',
                        'type': 'text',
                        'markdown': True,
                        'conditions': [
                            {
                                'type': 'choice_equals',
                                'property_name': 'dropdown',
                                'choice': {'en': 'en 1', 'de': 'de 1'}
                            },
                            {
                                'type': 'user_equals',
                                'property_name': 'user',
                                'user_id': None
                            },
                            {
                                'type': 'bool_equals',
                                'property_name': 'checkbox',
                                'value': True
                            },
                            {
                                'type': 'object_equals',
                                'property_name': 'object',
                                'object_id': 1
                            }
                        ]
                    }
                },
                'required': ['name']
            }
        )
        set_action_translation(Language.ENGLISH, action.id, name="Conditions Demo Action", description="")
        sampledb.logic.action_permissions.set_action_permissions_for_all_users(action.id, sampledb.models.Permissions.READ)
        user_reference_object = sampledb.logic.objects.create_object(
            action_id=action.id,
            data={
                'name': {
                    '_type': 'text',
                    'text': 'Object with a User Reference'
                },
                'user': {
                    '_type': 'user',
                    'user_id': basic_user.id
                }
            },
            user_id=instrument_responsible_user.id
        )

        UUID = '28b8d3ca-fb5f-59d9-8090-bfdbd6d07a71'
        component = sampledb.logic.components.add_component(UUID, 'Example SampleDB', None, 'Example component database for demonstration purposes. Do not expect it to function.')
        sampledb.logic.federation.components.parse_import_component_info({
            'uuid': 'e50671c5-2b31-4251-a24b-5b39bb6aa6e6',
            'source_uuid': UUID,
            'distance': 1,
            'name': None,
            'address': None,
            'discoverable': True
        }, component)
        sampledb.logic.federation.components.parse_import_component_info({
            'uuid': '9b9c5c43-4a7a-4a71-acc8-e75e706feaac',
            'source_uuid': UUID,
            'distance': 1,
            'name': 'Other SampleDB',
            'address': 'http://example.org',
            'discoverable': True
        }, component)
        sampledb.logic.federation.components.parse_import_component_info({
            'uuid': 'e50671c5-2b31-4251-a24b-5b39bb6aa6e6',
            'source_uuid': '9b9c5c43-4a7a-4a71-acc8-e75e706feaac',
            'distance': 2,
            'name': None,
            'address': None,
            'discoverable': True
        }, component)
        sampledb.logic.federation.locations.parse_import_location({
            'location_id': 1,
            'component_uuid': UUID,
            'name': {'en': 'Collaborating Institute', 'de': 'Partnerinstitut'},
            'description': {'en': 'A collaborating partner’s site.',
                            'de': 'Gelände eines Kooperationspartners.'},
        }, component)
        sampledb.logic.federation.locations.parse_import_location({
            'location_id': 2,
            'component_uuid': UUID,
            'name': {'en': 'Room 123', 'de': 'Raum 123'},
            'description': {'en': 'Room 123 on a collaborating partner’s site.',
                            'de': 'Raum 123 auf dem Gelände eines Kooperationspartners'},
            'parent_location': {'location_id': 1, 'component_uuid': UUID}
        }, component)
        sampledb.logic.federation.users.parse_import_user({
            'user_id': 1,
            'component_uuid': UUID,
            'name': 'Partnering User',
            'email': 'partner@example.com',
            'affiliation': 'Collaborating Partner LLC',
        }, component)
        sampledb.logic.federation.users.parse_import_user({
            'user_id': 2,
            'component_uuid': UUID,
            'name': None,
            'email': None,
            'affiliation': 'Collaborating Partner LLC',
        }, component)
        sampledb.logic.federation.instruments.parse_import_instrument({
            'instrument_id': 1,
            'component_uuid': UUID,
            'description_is_markdown': False,
            'short_description_is_markdown': False,
            'notes_is_markdown': False,
            'is_hidden': False,
            'translations': {'en': {'name': 'Collaborating Partner’s Measurement Instrument',
                                    'description': '',
                                    'short_description': '',
                                    'notes': ''},
                             'de': {'name': 'Messinstrument des Kooperationspartner',
                                    'description': '',
                                    'short_description': '',
                                    'notes': ''}}
        }, component)
        sampledb.logic.federation.action_types.parse_import_action_type({
            'action_type_id': ActionType.SAMPLE_CREATION,
            'component_uuid': UUID,
            'admin_only': False,
            'enable_labels': True,
            'enable_files': True,
            'enable_locations': True,
            'enable_publications': True,
            'enable_comments': True,
            'enable_activity_log': True,
            'enable_related_objects': True,
            'enable_project_link': False,
            'translations': {'en': {'name': 'Sample Creation',
                                    'description': 'These Actions represent processes which create a sample.',
                                    'object_name': 'Sample',
                                    'object_name_plural': 'Samples',
                                    'view_text': 'View Samples',
                                    'perform_text': 'Create Sample'},
                             'de': {'name': 'Probenerstellung',
                                    'description': 'Diese Aktionen repräsentieren Prozesse, die Proben erstellen.',
                                    'object_name': 'Probe',
                                    'object_name_plural': 'Proben',
                                    'view_text': 'Proben anzeigen',
                                    'perform_text': 'Probe erstellen'}}
        }, component)
        sampledb.logic.federation.actions.parse_import_action({
            'action_id': 1,
            'component_uuid': UUID,
            'action_type': {'action_type_id': ActionType.SAMPLE_CREATION, 'component_uuid': UUID},
            'description_is_markdown': False,
            'short_description_is_markdown': False,
            'instrument': None,
            'schema': {'title': 'Sampling', 'type': 'object', 'properties': {'name': {'title': 'Additional Note', 'type': 'text'}}, 'required': ['name']},
            'translations': {'en': {'name': 'Sampling',
                                    'description': 'Sample creation.',
                                    'short_description': 'Sample creation.'},
                             'de': {'name': 'Probenentnahme',
                                    'description': 'Probenentnahmeverfahren.',
                                    'short_description': 'Probenentnahmeverfahren.'}}
        }, component)
        sampledb.logic.federation.actions.parse_import_action({
            'action_id': 2,
            'component_uuid': UUID,
            'action_type': {'action_type_id': ActionType.MEASUREMENT, 'component_uuid': UUID},
            'description_is_markdown': False,
            'short_description_is_markdown': False,
            'instrument': {'instrument_id': 1, 'component_uuid': UUID},
            'schema': {'title': 'Special Measurement', 'type': 'object', 'properties': {'name': {'title': 'Additional Note', 'type': 'text'}, 'sample1': {'title': 'Sample 1', 'type': 'sample'}, 'sample2': {'title': 'Sample 2', 'type': 'sample'}}, 'required': ['name']},
            'translations': {'en': {'name': 'Special Measurement',
                                    'description': 'A special measurement only performed at the collaborating partner.',
                                    'short_description': 'A special measurement only performed at the collaborating partner.'},
                             'de': {'name': 'Spezielle Messung',
                                    'description': 'Eine spezielle Messung, die nur beim Kooperationspartner ausgeführt werden kann.',
                                    'short_description': 'Eine spezielle Messung, die nur beim Kooperationspartner ausgeführt werden kann.'}}
        }, component)
        sampledb.logic.federation.objects.parse_import_object({
            'object_id': 1,
            'versions': [{
                'version_id': 0,
                'data': {'name': {'_type': 'text', 'text': 'Shared Sample'}},
                'schema': {'title': 'Sampling', 'type': 'object', 'properties': {'name': {'title': 'Additional Note', 'type': 'text'}}, 'required': ['name']},
                'user': {'user_id': 2, 'component_uuid': UUID},
                'utc_datetime': datetime.datetime.now(datetime.timezone.utc).strftime('%Y-%m-%d %H:%M:%S.%f')
            }],
            'action': {'action_id': 1, 'component_uuid': UUID},
            'policy': {
                'access': {'data': True, 'files': True, 'action': True, 'comments': True, 'user_ids': True, 'user_data': True, 'object_location_assignments': True},
                'permissions': {'users': {basic_user.id: 'read'}, 'groups': {group_id: 'read'}, 'projects': {project_id: 'read'}}
            }
        }, component)
        sampledb.logic.federation.objects.parse_import_object({
            'object_id': 2,
            'versions': [{
                'version_id': 0,
                'data': {'name': {'_type': 'text', 'text': 'Shared Measurement'}, 'sample1': {'_type': 'sample', 'object_id': 1, 'component_uuid': UUID}, 'sample2': {'_type': 'sample', 'object_id': 3, 'component_uuid': UUID, 'export_edit_note': 'This internal sample was not exported.'}},
                'schema': {'title': 'Special Measurement', 'type': 'object', 'properties': {'name': {'title': 'Additional Note', 'type': 'text'}, 'sample1': {'title': 'Sample 1', 'type': 'sample'}, 'sample2': {'title': 'Sample 2', 'type': 'sample'}}, 'required': ['name']},
                'user': {'user_id': 1, 'component_uuid': UUID},
                'utc_datetime': datetime.datetime.now(datetime.timezone.utc).strftime('%Y-%m-%d %H:%M:%S.%f')
            }],
            'action': {'action_id': 1, 'component_uuid': UUID},
            'policy': {
                'access': {'data': True, 'files': True, 'action': True, 'comments': True, 'user_ids': True, 'user_data': True, 'object_location_assignments': True},
                'permissions': {'users': {basic_user.id: 'read'}, 'groups': {group_id: 'read'}, 'projects': {project_id: 'read'}}
            },
            'sharing_user': {
                'user_id': 1,
                'component_uuid': UUID
            },
            'comments': [
                {
                    'comment_id': 1,
                    'component_uuid': UUID,
                    'user': {'user_id': 1, 'component_uuid': UUID},
                    'content': 'I want to comment here.',
                    'utc_datetime': datetime.datetime.now(datetime.timezone.utc).strftime('%Y-%m-%d %H:%M:%S.%f')
                },
                {
                    'comment_id': 2,
                    'component_uuid': UUID,
                    'user': None,
                    'content': 'Another important comment by an anonymous user.',
                    'utc_datetime': datetime.datetime.now(datetime.timezone.utc).strftime('%Y-%m-%d %H:%M:%S.%f')
                },
                {
                    'comment_id': 3,
                    'component_uuid': UUID,
                    'user': {'user_id': 4, 'component_uuid': UUID},
                    'content': 'You might be interested in my specific comment regarding this object.',
                    'utc_datetime': datetime.datetime.now(datetime.timezone.utc).strftime('%Y-%m-%d %H:%M:%S.%f')
                }
            ],
            'object_location_assignments': [
                {
                    'id': 1,
                    'component_uuid': UUID,
                    'user': {'user_id': 1, 'component_uuid': UUID},
                    'responsible_user': {'user_id': 1, 'component_uuid': UUID},
                    'location': {'location_id': 1, 'component_uuid': UUID},
                    'description': {'en': ''},
                    'confirmed': True,
                    'utc_datetime': datetime.datetime.now(datetime.timezone.utc).strftime('%Y-%m-%d %H:%M:%S.%f')
                },
                {
                    'id': 2,
                    'component_uuid': UUID,
                    'user': {'user_id': 1, 'component_uuid': UUID},
                    'responsible_user': {'user_id': 3, 'component_uuid': UUID},
                    'location': {'location_id': 2, 'component_uuid': UUID},
                    'description': {'en': ''},
                    'confirmed': False,
                    'utc_datetime': datetime.datetime.now(datetime.timezone.utc).strftime('%Y-%m-%d %H:%M:%S.%f')
                }
            ],
            'files': [
                {
                    'file_id': 1,
                    'component_uuid': UUID,
                    'user': {'user_id': 3, 'component_uuid': UUID},
                    'data': {"storage": "url", "url": "https://example.com/file"},
                    'utc_datetime': datetime.datetime.now(datetime.timezone.utc).strftime('%Y-%m-%d %H:%M:%S.%f')
                }
            ]
        }, component)
        sampledb.logic.federation.objects.parse_import_object({
            'object_id': 3,
            'versions': [{
                'version_id': 0,
                'data': {'name': {'_type': 'text', 'text': 'Shared Object Without Shared Action'}},
                'schema': {'title': 'Sampling', 'type': 'object',
                           'properties': {'name': {'title': 'Additional Note', 'type': 'text'}}, 'required': ['name']},
                'user': {'user_id': 2, 'component_uuid': UUID},
                'utc_datetime': datetime.datetime.now(datetime.timezone.utc).strftime('%Y-%m-%d %H:%M:%S.%f')
            }],
            'action': {'action_id': 3, 'component_uuid': UUID},
            'policy': {
                'access': {'data': True, 'files': True, 'action': True, 'comments': True, 'user_ids': True,
                           'user_data': True, 'object_location_assignments': True},
                'permissions': {'users': {basic_user.id: 'read'}, 'groups': {group_id: 'read'},
                                'projects': {project_id: 'read'}}
            }
        }, component)
        sampledb.logic.federation.objects.parse_import_object({
            'object_id': 4,
            'versions': [{
                'version_id': 0,
                'data': {'name': {'_type': 'text', 'text': 'Shared Object with File References'}, 'file_a': {'_type': 'file', 'file_id': 1, 'component_uuid': component.uuid}, 'file_b': {'_type': 'file', 'file_id': 2, 'component_uuid': component.uuid}},
                'schema': {'title': 'Object Information', 'type': 'object', 'properties': {'name': {'title': 'Name', 'type': 'text'}, 'file_a': {'title': 'File A', 'type': 'file'}, 'file_b': {'title': 'File B', 'type': 'file'}}, 'required': ['name']},
                'user': {'user_id': 1, 'component_uuid': UUID},
                'utc_datetime': datetime.datetime.now(datetime.timezone.utc).strftime('%Y-%m-%d %H:%M:%S.%f')
            }],
            'action': {'action_id': 1, 'component_uuid': UUID},
            'policy': {
                'access': {'data': True, 'files': True, 'action': True, 'comments': True, 'user_ids': True, 'user_data': True, 'object_location_assignments': True},
                'permissions': {'users': {basic_user.id: 'read'}, 'groups': {group_id: 'read'}, 'projects': {project_id: 'read'}}
            },
            'sharing_user': {
                'user_id': 1,
                'component_uuid': UUID
            },
            'comments': [],
            'object_location_assignments': [],
            'files': [
                {
                    'file_id': 1,
                    'component_uuid': UUID,
                    'user': {'user_id': 1, 'component_uuid': UUID},
                    'data': {"storage": "federation", "original_file_name": "example.txt"},
                    'utc_datetime': datetime.datetime.now(datetime.timezone.utc).strftime('%Y-%m-%d %H:%M:%S.%f')
                }
            ]
        }, component)

        institute_category = sampledb.logic.group_categories.create_group_category(
            name={
                'de': 'Demo-Institut',
                'en': 'Demo Institute'
            }
        )

        lab_category = sampledb.logic.group_categories.create_group_category(
            name={
                'de': 'Demo-Labor',
                'en': 'Demo Laboratory'
            },
            parent_category_id=institute_category.id
        )
        sampledb.logic.group_categories.set_project_group_categories(project_id, [institute_category.id, lab_category.id])
        sampledb.logic.group_categories.set_project_group_categories(project_id2, [lab_category.id])
        sampledb.logic.group_categories.set_basic_group_categories(group_id, [institute_category.id])

        multi_measurement_action = sampledb.logic.actions.create_action(
            action_type_id=ActionType.MEASUREMENT,
            schema={
                'title': 'Example Object',
                'type': 'object',
                'properties': {
                    'name': {
                        'title': 'Object Name',
                        'type': 'text',
                        'languages': ['en', 'de']
                    },
                    'samples': {
                        'title': 'Samples',
                        'type': 'array',
                        'style': 'list',
                        'items': {
                            'title': 'Sample',
                            'type': 'sample'
                        }
                    },
                    'comment': {
                        'title': {'en': 'Comment', 'de': 'Kommentar'},
                        'type': 'text',
                        'markdown': True,
                        'languages': "all"
                    }
                },
                'required': ['name']
            }
        )
        set_action_translation(Language.ENGLISH, multi_measurement_action.id, name="multi measurement_action", description="")
        sampledb.logic.action_permissions.set_action_permissions_for_all_users(multi_measurement_action.id, sampledb.models.Permissions.READ)

        sampledb.logic.shares.add_object_share(
            object_id=measurement.id,
            component_id=component.id,
            policy={
                'access': {},
                'permissions': {
                    'users': {}
                }
            },
            user_id=basic_user.id
        )
        sampledb.logic.shares.update_object_share(
            object_id=measurement.id,
            component_id=component.id,
            policy={
                'access': {},
                'permissions': {
                    'users': {1: 'read'},
                    'groups': {2: 'write'},
                    'projects': {3: 'grant'}
                }
            },
            user_id=instrument_responsible_user.id
        )
        sampledb.logic.shares.set_object_share_import_status(
            object_id=measurement.id,
            component_id=component.id,
            import_status={
                'success': False,
                'notes': ['Demo import error'],
                'utc_datetime': datetime.datetime.now(datetime.timezone.utc).strftime('%Y-%m-%d %H:%M:%S'),
                'object_id': None
            }
        )
        sampledb.logic.shares.set_object_share_import_status(
            object_id=measurement.id,
            component_id=component.id,
            import_status={
                'success': True,
                'notes': ['Demo import note'],
                'utc_datetime': datetime.datetime.now(datetime.timezone.utc).strftime('%Y-%m-%d %H:%M:%S'),
                'object_id': 14
            }
        )
        sampledb.logic.shares.set_object_share_import_status(
            object_id=measurement.id,
            component_id=component.id,
            import_status={
                'success': True,
                'notes': [],
                'utc_datetime': datetime.datetime.now(datetime.timezone.utc).strftime('%Y-%m-%d %H:%M:%S'),
                'object_id': 14
            }
        )
        timeseries_action = sampledb.logic.actions.create_action(
            action_type_id=ActionType.MEASUREMENT,
            schema={
                'title': 'Example Object',
                'type': 'object',
                'properties': {
                    'name': {
                        'title': 'Object Name',
                        'type': 'text',
                        'languages': ['en', 'de']
                    },
                    'temperature_series': {
                        'title': 'Temperature Series',
                        'type': 'timeseries',
                        'units': ['degC', 'K'], 'display_digits': 2
                    },
                    'pressure_series': {
                        'title': 'Pressure Series',
                        'type': 'timeseries',
                        'units': 'bar',
                        'display_digits': 2,
                        'statistics': ['first', 'last', 'count']
                    }
                },
                'required': ['name']
            }
        )
        set_action_translation(Language.ENGLISH, timeseries_action.id, name="Timeseries Demo Action", description="")
        sampledb.logic.action_permissions.set_action_permissions_for_all_users(timeseries_action.id, sampledb.models.Permissions.READ)
        temperature_data: typing.List[typing.Tuple[str, float, float]] = []
        pressure_data: typing.List[typing.Tuple[str, float]] = []
        random.seed(0)
        for i in range(1000):
            utc_datetime_string = (datetime.datetime.now(datetime.timezone.utc) + datetime.timedelta(milliseconds=i * 30)).strftime('%Y-%m-%d %H:%M:%S.%f')
            if i == 0:
                temperature_magnitude = 20.0
            else:
                temperature_magnitude = temperature_data[-1][1] + random.uniform(-1, 1)
            temperature_magnitude_in_base_units = temperature_magnitude + 273.15
            temperature_data.append((
                utc_datetime_string,
                temperature_magnitude,
                temperature_magnitude_in_base_units
            ))
            if i == 0:
                pressure_magnitude = 2.0
            else:
                pressure_magnitude = pressure_data[-1][1] + random.uniform(-0.1, 0.1)
            pressure_data.append((
                utc_datetime_string,
                pressure_magnitude
            ))
        data = {
            'name': {
                '_type': 'text',
                'text': {'en': 'Timeseries Demo 1', 'de': 'Timeseries Demo 1'}
            },
            'temperature_series': {
                '_type': 'timeseries',
                'units': 'degC',
                'data': temperature_data
            },
            'pressure_series': {
                '_type': 'timeseries',
                'units': 'bar',
                'data': pressure_data
            }
        }
        sampledb.logic.objects.create_object(timeseries_action.id, data, instrument_responsible_user.id)

        relative_temperature_data: typing.List[typing.Tuple[float, float, float]] = [(i * 0.03, data[1], data[2]) for i, data in enumerate(temperature_data)]
        relative_pressure_data: typing.List[typing.Tuple[float, float]] = [(i * 0.03, data[1]) for i, data in enumerate(pressure_data)]
        data = {
            'name': {
                '_type': 'text',
                'text': {'en': 'Relative Timeseries Demo', 'de': 'Relative Timeseries Demo'}
            },
            'temperature_series': {
                '_type': 'timeseries',
                'units': 'degC',
                'data': relative_temperature_data
            },
            'pressure_series': {
                '_type': 'timeseries',
                'units': 'bar',
                'data': relative_pressure_data
            }
        }
        sampledb.logic.objects.create_object(timeseries_action.id, data, instrument_responsible_user.id)

        combined_conditions_action = sampledb.logic.actions.create_action(
            action_type_id=ActionType.MEASUREMENT,
            schema={
                'title': 'Example Object',
                'type': 'object',
                'properties': {
                    'name': {
                        'title': 'Object Name',
                        'type': 'text',
                        'languages': ['en', 'de']
                    },
                    'choice': {
                        'title': 'Choice',
                        'type': 'text',
                        'choices': ['A', 'B']
                    },
                    'checkbox_a': {
                        'title': 'Checkbox A',
                        'type': 'bool',
                        'conditions': [
                            {
                                'type': 'choice_equals',
                                'property_name': 'choice',
                                'choice': 'A'
                            }
                        ]
                    },
                    'checkbox_b': {
                        'title': 'Checkbox B',
                        'type': 'bool',
                        'conditions': [
                            {
                                'type': 'choice_equals',
                                'property_name': 'choice',
                                'choice': 'B'
                            }
                        ]
                    },
                    'text_a': {
                        'title': 'Text A',
                        'type': 'text',
                        'conditions': [
                            {
                                'type': 'bool_equals',
                                'property_name': 'checkbox_a',
                                'value': True
                            }
                        ]
                    },
                    'text_b': {
                        'title': 'Text B',
                        'type': 'text',
                        'conditions': [
                            {
                                'type': 'bool_equals',
                                'property_name': 'checkbox_b',
                                'value': True
                            }
                        ]
                    }
                },
                'required': ['name']
            }
        )
        set_action_translation(Language.ENGLISH, combined_conditions_action.id, name="Combined Conditions Action", description="")
        sampledb.logic.action_permissions.set_action_permissions_for_all_users(combined_conditions_action.id, sampledb.models.Permissions.READ)

        file_action = sampledb.logic.actions.create_action(
            action_type_id=ActionType.MEASUREMENT,
            schema={
                'title': 'Example Object',
                'type': 'object',
                'properties': {
                    'name': {
                        'title': 'Object Name',
                        'type': 'text',
                        'languages': ['en', 'de']
                    },
                    'file': {
                        'title': 'Example File',
                        'type': 'file',
                        'extensions': ['.txt', '.png'],
                        'preview': True
                    },
                    'list': {
                        'title': 'File List',
                        'type': 'array',
                        'style': 'list',
                        'items': {
                            'title': 'Example File',
                            'type': 'file',
                            'extensions': ['.txt', '.png']
                        }
                    },
                    'table': {
                        'title': 'File Table',
                        'type': 'array',
                        'style': 'table',
                        'items': {
                            'title': 'Example Object',
                            'type': 'object',
                            'properties': {
                                'file': {
                                    'title': 'Example File',
                                    'type': 'file',
                                    'extensions': ['.txt', '.png'],
                                    'preview': True
                                }
                            },
                            'required': ['file']
                        }
                    }
                },
                'required': ['name'],
                'propertyOrder': ['name', 'file', 'list', 'table'],
                'displayProperties': ['file']
            }
        )
        set_action_translation(Language.ENGLISH, file_action.id, name="File Action", description="")
        sampledb.logic.action_permissions.set_action_permissions_for_all_users(file_action.id, sampledb.models.Permissions.READ)

        file_object = sampledb.logic.objects.create_object(
            action_id=file_action.id,
            data={
                'name': {
                    '_type': 'text',
                    'text': {
                        'en': 'File Demo Object',
                        'de': 'Datei-Demo-Objekt',
                    }
                },
                'list': [],
                'table': []
            },
            user_id=basic_user.id
        )
        file = sampledb.logic.files.create_database_file(
            object_id=file_object.id,
            user_id=basic_user.id,
            file_name='example.png',
            save_content=lambda f: typing.cast(None, f.write(ghs01_image))
        )
        preview_image: Image.Image = Image.open(io.BytesIO(ghs01_image), formats=['PNG'])
        preview_image = preview_image.resize((10, 10), Image.Resampling.LANCZOS).resize((100, 100), Image.Resampling.NEAREST)
        preview_image_file = io.BytesIO()
        preview_image.save(preview_image_file, format='PNG')
        preview_image_file.seek(0)
        preview_image_binary_data = preview_image_file.read()
        sampledb.logic.files.create_database_file(
            object_id=file_object.id,
            user_id=basic_user.id,
            file_name='example.txt',
            save_content=lambda f: typing.cast(None, f.write(b'Dies ist ein Test')),
            preview_image_binary_data=preview_image_binary_data,
            preview_image_mime_type='image/png'
        )
        sampledb.logic.objects.update_object(
            object_id=file_object.id,
            data={
                'name': {
                    '_type': 'text',
                    'text': {
                        'en': 'File Demo Object',
                        'de': 'Datei-Demo-Objekt',
                    }
                },
                'file': {
                    '_type': 'file',
                    'file_id': file.id
                },
                'list': [],
                'table': []
            },
            user_id=basic_user.id
        )
        sampledb.logic.object_permissions.set_object_permissions_for_all_users(file_object.id, sampledb.models.Permissions.READ)

        sampledb.logic.comments.create_comment(
            object_id=independent_object.id,
            user_id=basic_user.id,
            content='This is a test comment',
            utc_datetime=datetime.datetime.now(datetime.timezone.utc) - datetime.timedelta(days=2),
        )
        sampledb.logic.files.create_url_file(
            object_id=independent_object.id,
            user_id=basic_user.id,
            url='http://example.org/test_url'
        )
        sampledb.logic.files.create_database_file(
            object_id=independent_object.id,
            user_id=basic_user.id,
            file_name='test_database.txt',
            save_content=lambda f: typing.cast(None, f.write(b"Test"))
        )
        with app.test_request_context('/'):
            eln_file_data = sampledb.logic.export.get_eln_archive(
                user_id=instrument_responsible_user.id,
                object_ids=[independent_object.id, file_object.id, measurement.id, user_reference_object.id]
            )
        eln_import_id = sampledb.logic.eln_import.create_eln_import(
            user_id=instrument_responsible_user.id,
            file_name='sampledb_demo_file.eln',
            zip_bytes=eln_file_data
        ).id
        sampledb.logic.eln_import.import_eln_file(eln_import_id)

        data = {
            'name': {
                '_type': 'text',
                'text': {
                    'en': 'Imported Object Reference Demo'
                }
            },
            'samples': [
                {
                    '_type': 'sample',
                    'object_id': 10001,
                    'component_uuid': 'ee36dd7f-72b0-44b6-afa8-752e920fbb32'
                },
                {
                    '_type': 'sample',
                    'object_id': 10001,
                    'eln_source_url': 'http://localhost:5000/',
                    'eln_object_url': 'http://localhost:5000/objects/10001'
                }
            ]
        }
        object = sampledb.logic.objects.create_object(multi_measurement_action.id, data, instrument_responsible_user.id)
        sampledb.logic.object_permissions.set_object_permissions_for_all_users(object.id, sampledb.models.Permissions.READ)

        choice_array_action = sampledb.logic.actions.create_action(
            action_type_id=ActionType.SAMPLE_CREATION,
            schema={
                'title': 'Example Object',
                'type': 'object',
                'properties': {
                    'name': {
                        'title': 'Object Name',
                        'type': 'text',
                        'languages': ['en', 'de']
                    },
                    'choice_array': {
                        'title': 'Choice Array',
                        'type': 'array',
                        'style': 'choice',
                        'items': {
                            'title': 'Choice Item',
                            'type': 'text',
                            'choices': [{'en': 'A'}, {'en': 'B'}, {'en': 'C'}, {'en': 'D (en)', 'de': 'D (de)'}]
                        }
                    }
                },
                'required': ['name'],
                'propertyOrder': ['name', 'choice_array']
            }
        )
        set_action_translation(Language.ENGLISH, choice_array_action.id, name="Choice Array Demo Action", description="")
        sampledb.logic.action_permissions.set_action_permissions_for_all_users(choice_array_action.id, sampledb.models.Permissions.READ)

        timeline_array_action = sampledb.logic.actions.create_action(
            action_type_id=ActionType.SAMPLE_CREATION,
            schema={
                "title": "Example Object",
                "type": "object",
                "properties": {
                    "name": {
                        "title": "Object Name",
                        "type": "text"
                    },
                    "timeline_array": {
                        "title": "Timeline Array",
                        "type": "array",
                        "style": "timeline",
                        "items": {
                            "title": "Event",
                            "type": "object",
                            "properties": {
                                "datetime": {
                                    "type": "datetime",
                                    "title": "Datetime"
                                },
                                "label": {
                                    "type": "text",
                                    "title": "Label", "languages": "all"
                                }
                            },
                            "required": ["datetime"]
                        }
                    }
                },
                "required": ["name"],
                "propertyOrder": ["name", "timeline_array"]
            }
        )
        set_action_translation(Language.ENGLISH, timeline_array_action.id, name="Timeline Array Demo Action", description="")
        sampledb.logic.action_permissions.set_action_permissions_for_all_users(timeline_array_action.id, sampledb.models.Permissions.READ)
        sampledb.logic.objects.create_object(
            action_id=timeline_array_action.id,
            data={
                "name": {
                    "_type": "text",
                    "text": {"en": "Timeline Array Demo Object"}
                },
                "timeline_array": [
                    {
                        "datetime": {
                            "_type": "datetime",
                            "utc_datetime": "2024-01-02 03:04:05"
                        }
                    },
                    {
                        "datetime": {
                            "_type": "datetime",
                            "utc_datetime": "2024-01-02 03:04:06"
                        },
                        "label": {
                            "_type": "text",
                            "text": "Very " * 5 + "Long Text"
                        }
                    },
                    {
                        "datetime": {
                            "_type": "datetime",
                            "utc_datetime": "2024-01-02 03:04:07"
                        },
                        "label": {
                            "_type": "text",
                            "text": {"en": "Translated Text EN", "de": "Translated Text DE"}
                        }
                    },
                    {
                        "datetime": {
                            "_type": "datetime",
                            "utc_datetime": "2024-01-02 03:04:04"
                        },
                        "label": {
                            "_type": "text",
                            "text": "HTML tags will be escaped<br />"
                        }
                    }
                ] + [
                    {
                        "datetime": {
                            "_type": "datetime",
                            "utc_datetime": "2024-01-02 03:04:05"
                        },
                        "label": {
                            "_type": "text",
                            "text": f"Event {i}"
                        }
                    }
                    for i in range(5)
                ]
            },
            user_id=instrument_responsible_user.id
        )
        large_table_action = sampledb.logic.actions.create_action(
            action_type_id=ActionType.SAMPLE_CREATION,
            schema={
                "title": {
                    "en": "Object Information"
                },
                "type": "object",
                "properties": {
                    "name": {
                        "title": {
                            "en": "Name"
                        },
                        "type": "text"
                    },
                    "table": {
                        "type": "array",
                        "title": {
                            "en": "Large Table"
                        },
                        "items": {
                            "type": "object",
                            "title": {
                                "en": "Row"
                            },
                            "properties": {
                                letter: {
                                    "title": {
                                        "en": f"Column {letter.upper()}"
                                    },
                                    "type": "text"
                                }
                                for letter in string.ascii_lowercase
                            }
                        },
                        "defaultItems": 1,
                        "style": {
                            "view": "full_width_table"
                        }
                    }
                },
                "required": [
                    "name"
                ],
                "propertyOrder": [
                    "name", "table"
                ]
            }
        )
        set_action_translation(Language.ENGLISH, large_table_action.id, name="Full Width Table Demo Action", description="")
        sampledb.logic.action_permissions.set_action_permissions_for_all_users(large_table_action.id, sampledb.models.Permissions.READ)

        collapsible_object_action = sampledb.logic.actions.create_action(
            action_type_id=ActionType.SAMPLE_CREATION,
            schema={
                "title": {
                    "en": "Object Information"
                },
                "type": "object",
                "properties": {
                    "name": {
                        "title": {
                            "en": "Name"
                        },
                        "type": "text"
                    },
                    "collapsible_object": {
                        "title": {
                            "en": "Collapsible Object"
                        },
                        "type": "object",
                        "style": "collapsible",
                        "note": "This object has the 'collapsible' style. It has a button to collapse it.",
                        "properties": {
                            "text_in_object": {
                                "title": {
                                    "en": "Text in Object"
                                },
                                "type": "text"
                            }
                        },
                        "propertyOrder": [
                            "text_in_object"
                        ]
                    },
                    "expandable_object": {
                        "title": {
                            "en": "Expandable Object"
                        },
                        "type": "object",
                        "style": "expandable",
                        "note": "This object has the 'expandable' style. It starts out collapsed with a button to expand it.",
                        "properties": {
                            "text_in_object": {
                                "title": {
                                    "en": "Text in Object"
                                },
                                "type": "text"
                            }
                        },
                        "propertyOrder": [
                            "text_in_object"
                        ]
                    }
                },
                "required": [
                    "name",
                    "collapsible_object",
                    "expandable_object"
                ],
                "propertyOrder": [
                    "name",
                    "collapsible_object",
                    "expandable_object"
                ]
            }
        )
        set_action_translation(Language.ENGLISH, collapsible_object_action.id, name="Collapsible/Expandable Object Action", description="")
        sampledb.logic.action_permissions.set_action_permissions_for_all_users(collapsible_object_action.id, sampledb.models.Permissions.READ)

        horizontal_object_action = sampledb.logic.actions.create_action(
            action_type_id=ActionType.SAMPLE_CREATION,
            schema={
                "title": {
                    "en": "Object Information"
                },
                "type": "object",
                "properties": {
                    "name": {
                        "title": {
                            "en": "Name"
                        },
                        "type": "text"
                    },
                    "quantity_object": {
                        "title": {
                            "en": "Quantity Object"
                        },
                        "type": "object",
                        "style": "horizontal",
                        "properties": {
                            "value": {
                                "title": {
                                    "en": "Value"
                                },
                                "type": "quantity",
                                "units": "kg"
                            },
                            "uncertainty": {
                                "title": {
                                    "en": "Uncertainty"
                                },
                                "type": "quantity",
                                "units": "kg"
                            }
                        },
                        "required": [
                            "value", "uncertainty"
                        ],
                        "propertyOrder": [
                            "value", "uncertainty",
                        ]
                    },
                    "large_object": {
                        "title": {
                            "en": "Large Object"
                        },
                        "type": "object",
                        "style": "horizontal",
                        "properties": {
                            property_name: {
                                "title": {
                                    "en": f"Property {property_title}"
                                },
                                "type": "text",
                                "languages": "all",
                                "note": f"This is a note for property {property_title}"
                            }
                            for property_name, property_title in zip(string.ascii_lowercase, string.ascii_uppercase)
                        }
                    }
                },
                "required": [
                    "name"
                ],
                "propertyOrder": [
                    "name", "quantity_object", "large_object"
                ]
            }
        )
        set_action_translation(Language.ENGLISH, horizontal_object_action.id, name="Horizontal Object Demo Action", description="")
        sampledb.logic.action_permissions.set_action_permissions_for_all_users(horizontal_object_action.id, sampledb.models.Permissions.READ)
        sampledb.logic.objects.create_object(
            action_id=horizontal_object_action.id,
            data={
                "name": {
                    "_type": "text",
                    "text": {"en": "Horizontal Object Demo Object"}
                },
                "quantity_object": {
                    "value": {
                        "_type": "quantity",
                        "units": "kg",
                        "magnitude": 1
                    },
                    "uncertainty": {
                        "_type": "quantity",
                        "units": "kg",
                        "magnitude": 0.2
                    }
                },
                "large_object": {
                    property_name: {
                        "_type": "text",
                        "text": {"en": "Value for " + property_name}
                    }
                    for property_name in string.ascii_lowercase
                }
            },
            user_id=instrument_responsible_user.id
        )
        horizontal_object_action_schema = copy.deepcopy(typing.cast(typing.Dict[str, typing.Any], horizontal_object_action.schema))
        horizontal_object_action_schema['displayProperties'] = ['name']
        sampledb.logic.actions.update_action(
            action_id=horizontal_object_action.id,
            schema=horizontal_object_action_schema,
        )

    print("Success: set up demo data", flush=True)



# coding: utf-8
"""
RESTful API for SampleDB
"""
import typing

import flask

from .authentication import multi_auth
from ..utils import Resource, ResponseData
from ...logic import errors, locations, location_permissions, utils
from ...models import Permissions

__author__ = 'Florian Rhiem <f.rhiem@fz-juelich.de>'


def location_to_json(location: locations.Location) -> typing.Dict[str, typing.Any]:
    return {
        'location_id': location.id,
        'name': utils.get_translated_text(location.name, 'en') if location.name else None,
        'description': utils.get_translated_text(location.description, 'en'),
        'parent_location_id': location.parent_location_id,
        'type_id': location.type_id,
        'is_hidden': location.is_hidden,
        'enable_object_assignments': location.enable_object_assignments,
    }


def location_type_to_json(location_type: locations.LocationType) -> typing.Dict[str, typing.Any]:
    return {
        'location_type_id': location_type.id,
        'name': location_type.name.get('en', None) if location_type.name else None,
    }


def object_location_assignment_to_json(object_location_assignment: locations.ObjectLocationAssignment) -> typing.Dict[str, typing.Any]:
    return {
        'object_id': object_location_assignment.object_id,
        'location_id': object_location_assignment.location_id,
        'responsible_user_id': object_location_assignment.responsible_user_id,
        'user_id': object_location_assignment.user_id,
        'description': utils.get_translated_text(object_location_assignment.description, 'en'),
        'utc_datetime': object_location_assignment.utc_datetime.strftime('%Y-%m-%d %H:%M:%S') if object_location_assignment.utc_datetime is not None else None
    }


class Location(Resource):
    @multi_auth.login_required
    def get(self, location_id: int) -> ResponseData:
        try:
            location = locations.get_location(location_id=location_id)
        except errors.LocationDoesNotExistError:
            return {
                "message": f"location {location_id} does not exist"
            }, 404
        permissions = location_permissions.location_permissions.get_permissions_for_user(
            resource_id=location_id,
            user_id=flask.g.user.id,
            include_all_users=True,
            include_groups=True,
            include_projects=True,
            include_admin_permissions=True,
            limit_readonly_users=True
        )
        if Permissions.READ not in permissions:
            return flask.abort(403)
        return location_to_json(location)


class Locations(Resource):
    @multi_auth.login_required
    def get(self) -> ResponseData:
        return [
            location_to_json(location)
            for location in location_permissions.get_locations_with_user_permissions(
                user_id=flask.g.user.id,
                permissions=Permissions.READ
            )
        ]


class LocationType(Resource):
    @multi_auth.login_required
    def get(self, location_type_id: int) -> ResponseData:
        try:
            location_type = locations.get_location_type(location_type_id=location_type_id)
        except errors.LocationTypeDoesNotExistError:
            return {
                "message": f"location type {location_type_id} does not exist"
            }, 404
        return location_type_to_json(location_type)


class LocationTypes(Resource):
    @multi_auth.login_required
    def get(self) -> ResponseData:
        return [
            location_type_to_json(location_type)
            for location_type in locations.get_location_types()
        ]


class ObjectLocationAssignment(Resource):
    @multi_auth.login_required
    def get(self, object_id: int, object_location_assignment_index: int) -> ResponseData:
        try:
            object_location_assignments = locations.get_object_location_assignments(object_id=object_id)
        except errors.ObjectDoesNotExistError:
            return {
                "message": f"object {object_id} does not exist"
            }, 404
        if object_location_assignment_index < 0 or object_location_assignment_index >= len(object_location_assignments):
            return {
                "message": f"location assignment {object_location_assignment_index} for object {object_id} does not exist"
            }, 404
        return object_location_assignment_to_json(object_location_assignments[object_location_assignment_index])


class ObjectLocationAssignments(Resource):
    @multi_auth.login_required
    def get(self, object_id: int) -> ResponseData:
        try:
            return [
                object_location_assignment_to_json(object_location_assignment)
                for object_location_assignment in locations.get_object_location_assignments(object_id)
            ]
        except errors.ObjectDoesNotExistError:
            return {
                "message": f"object {object_id} does not exist"
            }, 404
